"""Tools for parsing various formats of cross section and
transport parameter data."""

import io
import itertools
import os
from os import listdir as ls
from os.path import join as pjoin
import re
import time
import warnings

import pandas as pd
import numpy as np

import matplotlib.pyplot as plt

def read_LXCat_cs(efile):
    """Convert LXCat cross section text files into
    python pandas DataFrames."""
    # Process types
    ptypes = ["ELASTIC", "EFFECTIVE", "ROTATIONAL",
              "EXCITATION", "IONIZATION", "ATTACHMENT"]
    # Some conditions indicating the beginning end of sections in the file
    # End of file
    EOF = lambda l: l == ""
    # Data base block start and exit
    db_start = lambda l: "DATABASE" in l
    db_end = lambda l: "x"*10 in l
    # Target block start and exit
    tg_start = lambda l: "*"*58 + " " in l
    tg_end = lambda l: "*"*80 in l
    # Process block start
    p_start = lambda l: l.rstrip() in ptypes
    def goto(l, *conditions):
        """Skip lines in string stream until a condition is met."""
        while not any(c(l) for c in conditions):
            l = f.readline()
        return l

    def read_L3(d, l):
        """Interpret the special third line of the LXCat format."""
        s = l.split()
        if len(s) > 1 and ": " not in l:
            # This must be an excitation with a population ratio
            d["threshold"] = float(s[0])
            d["weight_ratio"] = float(s[1])
            l = f.readline()
        elif d["process_type"] == "ATTACHMENT":
            # 3rd line is omitted
            pass
        elif len(s) > 1 and ": " in l:
            # 3rd line is missing for some reason
            pass
        elif d["process_type"] in ["ELASTIC", "EFFECTIVE"]:
            # 3rd line is the ratio between the electron and target mass
            d["target_mass_ratio"] = float(l.rstrip())
            l = f.readline()
        else:
            # 3rd line is the threshold value
            d["threshold"] = float(l.rstrip())
            l = f.readline()
        return l

    def scan_process(l):
        """Scan one block representing a single cross section table.
        Assume `l` begins at a `process_type` line in the LXCat file.
        """
        d = {"process_type": l.rstrip()}
        l = f.readline()
        d["process_string"] = l.rstrip()
        l = f.readline()
        l = read_L3(d, l)
        while "----" not in l:
            meta_data = l.split(": ")
            key = meta_data[0]
            msg = ": ".join(meta_data[1:])
            d[key.lower()] = msg.strip()
            l = f.readline()
        if "columns" not in d:
            d["columns"] = ["Energy (eV)", "Cross section (m2)"]
        else:
            d["columns"] = d["columns"].split(" | ")
        l = f.readline()
        es = []
        css = []
        while "-"*5 not in l:
            e, cs = map(float, l.split())
            es.append(e)
            css.append(cs)
            l = f.readline()
        # Add metadata
        df = pd.DataFrame(np.array([es, css]).T, columns=d["columns"])
        for k, v in d.items():
            if k != "columns":
                df[k] = v
        return df, l

    df = pd.DataFrame()
    with open(efile, "r") as f:
        l = f.readline()
        while not EOF(l):
            l = goto(l, db_start)
            db = l.split()[1]
            l = goto(l, tg_start)
            while not db_end(l):
                target = l.split()[1]
                l = goto(l, p_start)
                while not (tg_end(l) or db_end(l)):
                    pdf, l = scan_process(l)
                    # Update target and db info, add to df
                    pdf["target"] = target
                    pdf["db"] = db
                    df = pd.concat((df, pdf), ignore_index=True)
                    l = goto(l, p_start, tg_end, db_end)
                l = goto(l, tg_start, db_end)
            l = goto(l, db_start, EOF)
    return df

def get_cs_defaults(infile):
    infile = "thunderboltz/input.in"
    cols = ["fid", "r1", "r2", "rtype", "B", "p1", "p1"]
    with open(infile, "r") as f:
        df = pd.DataFrame()
        for line in f.readlines():
            if "CS" not in line or "CC" in line:
                continue
            l = line.split()[1:]
            fid = l[0].split("/")[-1].split(".")[0]
            dat = [fid, int(l[1]), int(l[2]), l[3], float(l[4]),
                   int(l[5]), int(l[6])]
            df = df.append({k: v for k, v in zip(cols, dat)},
                ignore_index=True)

def read_LXCat_swarmp(efile):
    """Read an LXCat swarm parameter text file and return a DataFrame."""
    df = pd.DataFrame()
    with open(efile, "r") as f:
        # For every db
        l = f.readline()
        while l:
            # Find first db
            maxout = 0
            while "DATABASE:" not in l and maxout < 1000:
                maxout += 1
                l = f.readline()
            if maxout > 999:
                break
            db = l.split()[1].upper()
            # For every author
            while "xxxxxx" not in l:
                while "SPECIES" not in l:
                    l = f.readline()
                conditions = T = None
                sp = " ".join(l.split()[1:])
                while "COMMENT" not in l:
                    l = f.readline()
                auth = " ".join(l.split()[1:])
                while "COLUMNS" not in l:
                    l = f.readline()
                x_col = l.split(":")[1].split("|")[0].strip()
                y_col = l.split(":")[1].split("|")[1].strip()
                l = f.readline()
                l = f.readline()
                x, y = [], []
                while "---" not in l:
                    if "UPDATED" in l:
                        continue
                    x_,  y_ = list(map(float, l.split()))
                    x.append(x_)
                    y.append(y_)
                    l = f.readline()
                n = len(x)
                dbs = [db]*n
                auths = [auth]*n
                conds = [conditions]*n
                Ts = [T]*n
                df2= pd.DataFrame({"db": dbs, "author": auths,
                    x_col: x, y_col: y, "conditions": conds,
                    "T": Ts})
                df = df.append(df2,ignore_index=True, sort=True)
                while ("xxxxxx" not in l and "SPECIES"not in l):
                    l = f.readline()
    return df


def extract_vdf(filename, sample_cap=500000, v=False):
    """Read a single particle dump file and return a DataFrame.

    Args:
        filename (str): The directory to search for particle dump files.
        sample_cap (int): Limit the number of samples read from the dump
            file for very large files. Default is 500000. If bool(sample_cap)
            evaluates to ``False``, then no cap will be imposed.
        v (int): Verbosity -- 0: silent, 1: print file paths before reading.

    Returns:
        :class:`pandas.DataFrame`: The velocity distributions function sample
            data.
    """
    if v: print(f"Reading {filename}", flush=True)

    # Read data
    with open(filename) as f:
        if sample_cap:
            # Read the first `sample_cap` lines if available
            txt = "".join(itertools.islice(f, int(sample_cap)))
        else:
            txt = "".join(f.readlines())

    try:
        vdf_arr = np.genfromtxt(
            io.StringIO(txt), delimiter=",", autostrip=True)
    except ValueError as E:
        warnings.warn(f"Skipping {efile}: {str(E)}", RuntimeWarning)
        return

    # Build up dataframe
    cols = ["vx", "vy", "vz"]
    vdf_df = pd.DataFrame(vdf_arr, columns=cols)
    return vdf_df

def extract_vdfs(calc_dir, steps="all", sample_cap=500000,
        particle_type=0, v=0):
    """
    Pull all electron velocity distribution data from a calculation directory
    into a DataFrame.

    Args:
        calc_dir (str): The directory to search for particle dump files.
        steps (list[int] or str): What time steps to extract from the dump files.
            If a list is passed, the files with corresponding step integers
            contained both within the list and the calculation directory
            will be extracted.
        sample_cap (int): Limit the number of samples read from the dump
            file for very large files. Default is 500000. If bool(sample_cap)
            evaluates to ``False``, then no cap will be imposed.
        v (int): Verbosity -- 0: silent, 1: print file paths before reading.

    Returns:
        :class:`pandas.DataFrame`: The velocity distributions function sample
            data.
    """
    # Initialize empty data frame
    df = pd.DataFrame({k: pd.Series([], dtype=dtype)
        for k, dtype in zip(["ptype", "step", "vx", "vy", "vz"],
                            [np.int64]*2 + [np.float64]*3)})

    if steps == []: return df # Return nothing, don't read anything

    if isinstance(particle_type, int):
        particle_type = [particle_type]

    for efile in ls(calc_dir):
        if "Dump" in efile: # Only read velocity dump files

            # Extract particle type and time step out of each filename.
            toks = efile.split("_")
            ptype = int(toks[1])
            step = int(toks[-1].split(".")[0])

            # Skip unrequested data
            if steps != "all" and step not in steps:
                continue
            if particle_type != "all" and ptype not in particle_type:
                continue

            vdf_df = extract_vdf(
                pjoin(calc_dir, efile), sample_cap=sample_cap, v=v)

            if vdf_df is None: continue

            vdf_df["step"] = step
            vdf_df["ptype"] = ptype
            df = pd.concat((df, vdf_df), ignore_index=True)

    df = df.sort_values("step").reset_index(drop=True)
    return df



import os

# Specify internal package data location
data_path = os.path.join(os.path.dirname(__file__), "data")
# Specify internal thunderboltz C++ source directory
src_path = os.path.join(os.path.dirname(__file__), "cpp")

# Allow main objects to be available on the package level 
from .tb import ThunderBoltz
from .tb import read
from .tb import query_tree
from .input import CrossSections
from .kinetic import Process





class Parameters(object):
    """An enumeration of parameters which can be indexed
    like a dictionary."""

    def __index__(self, s):
        return getattr(self, s)

    def get_params(self):
        """Return the set of parameters and their default values
        as a python dictionary."""
        return {k: v for k, v in self.__class__.__dict__.items()
                if "__" not in k}

class TBParameters(Parameters):
    """The ThunderBoltz simulation settings and their
    default values."""
    #: (float) Cell length (m), default is ``1e-6``.
    L=1e-6
    #: (int) Number of species, default is ``2``.
    SP=2
    #: (list[int]) Number of particles for each species,
    #: default is ``[10000, 1000]``.
    NP=[10000, 1000]
    #: (list[float]) Temperature (eV) of each particle species, default
    #: is ``[0.0, 0.0259]``.
    TP=[0.0, 0.0259]
    #: (list[float]) Flow velocity for each particle, default is ``[0.0, 0.0]``.
    VV=[0.0, 0.0]
    #: (list[int]) Charge (elementary units) of each particle species,
    #: default is ``[-1.0, 0.0]``.
    QP=[-1.0, 0.0]
    #: (list[float]) Mass of each particle species (amu),
    #: default is ``[5.4857e-4, 28.0]``.
    MP=[5.4857e-4, 28.0]
    #: (int) Number of time steps, default is ``1000001``.
    NS=1000001
    #: (int) Number of random samples used to find
    #: :math:`\max_\epsilon (v\sigma(\epsilon))` for each process,
    #: default is ``1000``.
    VS=1000
    #: (float) Time increment interval (s), default is ``5e-12``.
    DT=5.0e-12
    #: (float) Electric field in z-direction (V/m), default is
    #: ``-24640.0``.
    E=-24640.0
    #: (float) E-field oscillation frequency (Hz), default is ``0``.
    ET=0.
    #: (list[int]) Magnetic field vector (Tesla),
    #: default is ``[0.0, 0.0, 0.0]``.
    B=[0.0, 0.0, 0.0]
    #: (list[int]) Output velocity dump settings [start, stride, species ID],
    #: default is ``[1000, 1000000, 0]``.
    FV=[1000, 1000000, 0]
    #: (int) Cross section extrapolation — options are ``0`` (extrapolated
    #: cross sections are set to :math:`0` m :math:`^2`), or
    #: ``1`` (linearly extrapolated from last two points), default is ``0``.
    EX=0
    #: (float) Request memory (GB) for particle arrays. If 0 is passed,
    #: then enough space for :math:`4N_{\rm init}` will be
    #: allocated, default is ``0``.
    MEM=0.
    #: (int) When using a SLURM manager on HPC, auto dump particle velocity
    #: data before job allocation runs out —
    #: options are ``0`` (don't auto dump) | ``1`` (dump using SLURM setup).
    SE=0
    #: (int) Time step stride for output parameters, default is ``100``.
    OS=100
    #: (int) If ``1``, then the remainder of :math:`N_{\rm pairs}` is
    #: carried into the next :math:`N_{\rm pairs}` evaluation of the
    #: same process, default is ``1``.
    CR=1
    #: (list[str,int]) Optionally load particle velocities from a comma separated
    #: text file, default is ``None``; specify the name of the file at index
    #: ``0`` and the particle species index it applies to at index ``1``.
    LV=None

class WrapParameters(Parameters):
    """Additional Python interface settings and
    their defaults. An asterisk (*) at the beginning
    of the description indicates a parameter that is
    specific to the built-in He model.
    """
    #: (float) Specify reduced electric field (Td),
    #: default is ``None``.
    Ered=None
    #: (list[float]) Specify reduced magnetic field (Hx),
    #: default is ``None``.
    Bred=None
    #: (int) The index of the neutral gas species (if applicable),
    #: default is ``1``.
    gas_index=1
    #: (float) Run simulation until duration (s) is complete,
    #: default is ``None``.
    duration=None
    #: (int) Number of background neutral gas particles
    #: (assumed gas species is of index 1), default is ``None``.
    NN=None
    #: (float) Set the ratio :math:`\frac{N_{\rm e}}{N_{\rm gas}}`,
    #: default is ``None``.
    pct_ion=None
    #: (float) Minimum number of pseudo pairs to be generated by the
    #: smallest cross section of interest, default is ``1.0``.
    Nmin=1.0
    #: (float) The initial energy (eV) of a hypothetical electron per time step,
    #: default is ``10``.
    EP_0=10.
    #: (float) The maximum change in energy (eV) of a hypothetical electron per time
    #: step, default is ``0.1``.
    DE=0.1
    #: (bool) Flag to calculate ``DT`` / ``NP`` / ``E`` from ``Ered`` / ``L`` /
    #: ``pct_ion`` / ``DE`` / ``EP_0``, default is ``False``.
    autostep=False
    #: (dict[str->float]) Multiply any property involved in the :math:`N_{\rm min}`
    #: constraint by a constant after the constraint is imposed; useful for
    #: convergence testing, default is ``{}`` (do nothing).
    pc_scale={}
    #: (callable or str) Function for auto-generating indeck and CS object or
    #: string path to directory with CS data and indeck,
    #: default is ``None``.
    indeck=None
    #: (list[str or int or 2D-array, int]) Initialize particles with velocity data —
    #: a string at index ``0`` will read velocity data from that file path;
    #: an int at index ``0`` will attempt to reinitialize a previous calculation
    #: from a that time step if that step dump file is available;
    #: an array or DataFrame of shape (NP,3) at index ``0``
    #: will create a velocity initialization file with the provided data,
    #: the value at index 1 should represent the species type.
    vdf_init=None
    #: (bool) If specified with ``vdf_init``, truncate the init file
    #: such that NP is satisfied, default is ``False``.
    downsample=False
    # Cross section model
    #: *(int) For the built-in ``thunderboltz.input.He_TB`` Helium model, include CCC excitation
    #: processes from the ground state to (up to and including) states with
    #: principle quantum number ``n``.
    n=3
    #: *(str or bool) For the built-in ``thunderboltz.input.He_TB`` Helium model, use
    #: either tabulated data, analytic fits, or a mix of both,
    #: options are ``False`` | ``True`` | ``"mixed"``.
    analytic_cs="mixed"
    #: *(int) For the built-in ``thunderboltz.input.He_TB``, this specifies
    #: the number of tabulated cross section values for analytic sampling.
    nsamples=250
    #: *(float) If ``analytic_cs`` is ``"mixed"``, use numerical data
    #: at energies lower than this threshold value (in eV), and use analytic
    #: data at higher energies.
    mix_thresh=300
    #: (bool) Flag to append "FixedParticle2" to each of the reaction types in the indeck
    fixed_background=True
    #: (str) Elastic angular distribution function model —
    #: options are ``"default"``, or ``"He_Park"``.
    eadf="default"
    #: *(str or None) The total elastic cross section model for the He built-in —
    #: options are ``"ICS"`` or ``"MTCS"``, default is ``"ICS"`` if an anisotropic
    #: angular distribution function is used and ``"MTCS"`` if an isotropic
    #: angular distribution function is used.
    ECS=None
    # Ionization model
    egen=False #: *(bool) Allow secondary electron generation for the ionization model.
    #: (str) Electron energy sharing distribution model — options are
    #: ``"default"`` (one takes all) | ``"equal"`` | ``"uniform"``.
    eesd="default"

class OutputParameters(Parameters):
    """A listing of the main output parameters of the simulation, these keywords
    are the named columns of the time series and steady state data frames
    returned by :meth:`~.thunderboltz.ThunderBoltz.get_timeseries` and
    :meth:`~.thunderboltz.ThunderBoltz.get_ss_params` respectively. These data tables also
    include the :class:`ParticleParameters` of the species at index :math:`0`.
    The steady state parameters returned by :meth:`~.thunderboltz.ThunderBoltz.get_ss_params`
    will also include standard deviations for each parameter indicated by an
    added "_std" suffix.
    """

    #: (float) The mean energy (eV) of the species at index :math:`0` (usually electrons),
    #: computed as
    #: :math:`\langle\epsilon\rangle = \frac{m_0}{2N_0}\sum_{i=1}^{N_0}v_{0i}^2`
    #: where :math:`m_0` and :math:`N_0` are the mass and particle count
    #: of the :math:`0^{\rm th}` species, and :math:`v_{0i}` is the velocity
    #: vector of the :math:`i^{\rm th}` particle of species :math:`0`.
    MEe=0.

    #: (float) The reduced flux mobility :math:`(\text{V}^{-1}\text{m}^{-1}\text{s}^{-1})`
    #: of the species at index :math:`0`, computed as
    #: :math:`\mu^fn_{\rm gas}=\frac{n_{\rm gas}}{E}\frac{1}{N_0}\sum_{i=1}^{N_0}v_{\parallel,0i}`
    #: where :math:`N_0` is the particle count of the :math:`0^{\rm th}` species, :math:`E` is
    #: field magnitude, :math:`n_{\rm gas}` is the density of the background
    #: gas, and :math:`v_{\parallel,0i}` is the velocity component
    #: parallel to the E field vector of the :math:`i^{\rm th}` th particle of species :math:`0`.
    mobN=0.

    #: (float) The reduced bulk mobility :math:`(\text{V}^{-1}\text{m}^{-1}\text{s}^{-1})`
    #: of the species at index :math:`0`, computed as
    #: :math:`\mu^b n_{\rm gas}=\frac{n_{\rm gas}}{E}\frac{d}{dt}\left[\frac{1}{N_0}\sum_{i=1}^{N_0}r_{\parallel,0i}\right]`
    #: where :math:`N_0` is the particle count of the :math:`0^{\rm th}` species, :math:`E` is
    #: field magnitude, :math:`n_{\rm gas}` is the density of the background gas,
    #: and :math:`r_{\parallel,0i}` is displacement component
    #: parallel to the E field vector of the :math:`i^{\rm th}` th particle of species :math:`0`.
    mobN_bulk=0.

    #: (float) The reduced flux Townshend ionization coefficient :math:`(\text{m}^2)`
    #: of the species at index :math:`0`, computed as
    #: :math:`\frac{\alpha^f}{n_{\rm gas}} = \frac{1}{n_0 n_{\rm gas}}\frac{dC_{\rm ion}}{dt}\times \left( \frac{1}{N_0}\sum_{i=1}^{N_0}v_{\parallel,0i}\right)^{-1}`
    #: where :math:`N_0` is the particle count of the :math:`0^{\rm th}` species, :math:`C_{\rm ion}` is
    #: the count of ionization events, :math:`n_0` and :math:`n_{\rm gas}` are the
    #: :math:`0^{\rm th}` and background gas densities repectively, and :math:`v_{\parallel,0i}` is velocity component
    #: parallel to the E field vector of the :math:`i^{\rm th}` th particle of species :math:`0`.
    a_n=0.

    #: (float) The reduced bulk Townshend ionization coefficient :math:`(\text{m}^2)`
    #: of the species at index :math:`0`, computed as
    #: :math:`\frac{\alpha^b}{n_{\rm gas}} = \frac{1}{n_0 n_{\rm gas}}\frac{dC_{\rm ion}}{dt}\times\left( \frac{d}{dt} \left[\frac{1}{N_0}\sum_{i=1}^{N_0}r_{\parallel,0i}\right] \right)^{-1}`,
    #: where :math:`N_0` is the particle count of the :math:`0^{\rm th}` species, :math:`C_{\rm ion}` is
    #: the count of ionization events, :math:`n_0` and :math:`n_{\rm gas}` are the
    #: :math:`0^{\rm th}` and background gas densities respectively, and :math:`r_{\parallel,0i}` is displacement component
    #: parallel to the E field vector of the :math:`i^{\rm th}` th particle of species :math:`0`.
    a_n_bulk=0.

    #: (float) :math:`n_{\rm gas}`, the number density of the background gas.
    n_gas=0.

    #: (float) The electric field component (V/m) in the :math:`z` direction, which
    #: can change in AC scenarios.
    E=0.

    t=0. #: (float) The time (s) elapsed in the simulation.

    #: (int) The number of time steps elapsed in the simulation, with
    #: ``t`` :math:`=0` corresponding to ``step`` :math:`=0`, and with
    #: ``t`` = ``DT`` corresponding to ``step`` :math:`=1`.
    step=0

    #: (float) The ionization rate coefficient, computed as
    #: :math:`\frac{1}{n_0n_{\rm gas}} \frac{dC_{\rm ion}}{dt}`.
    k_ion=0.

    #: (float) The rate coefficient for the first reaction, computed as
    #: :math:`\frac{1}{n_0n_{\rm gas}} \frac{dC_{\rm 1}}{dt}`. If there are
    #: :math:`n` reactions included in the model, there will be the
    #: corresponding rate coefficients, :math:`k_1, k_2, ..., k_n` also
    #: included in the output tables.
    k_1=0.

class ParticleParameters(Parameters):
    """A listing of species dependent properties that can be accessed by
    :meth:`~.thunderboltz.ThunderBoltz.get_particle_tables`, which returns a
    list of data tables (one for each species) where each column of data is
    labeled with one of the following keywords."""
    Rxi=0. #: (float) The mean x component of all particle displacements (m).
    Ryi=0. #: (float) The mean y component of all particle displacements (m).
    Rzi=0. #: (float) The mean z component of all particle displacements (m).
    Vxi=0. #: (float) The mean x component velocity (m/s).
    Vyi=0. #: (float) The mean y component velocity (m/s).
    Vzi=0. #: (float) The mean z component velocity (m/s).
    Txi=0. #: (float) The mean x component temperature (eV).
    Tyi=0. #: (float) The mean y component temperature (eV).
    Tzi=0. #: (float) The mean z component temperature (eV).
    Ni=0.0 #: (float) The number density (m :math:`^{-3}`).
    Ki=0.0 #: (float) The total kinetic energy (eV).
    Mi=0.0 #: (float) The mean kinetic energy (eV).
    t=0. #: (float) The time (s) elapsed in the simulation.
    #: (int) The number of time steps elapsed in the simulation, with
    #: ``t`` :math:`=0` corresponding to ``step`` :math:`=0`.
    step=0

tb_parameters = TBParameters().get_params()
wrap_parameters = WrapParameters().get_params()
output_parameters = OutputParameters().get_params()
particle_parameters = ParticleParameters().get_params()



"""The main ThunderBoltz wrapper object with relevant helper functions."""
import copy
from datetime import datetime
import io
import json
import multiprocessing as mp
import os
from os import listdir as ls
from os.path import join as pjoin
from pathlib import Path
import re
import shutil
import subprocess
from subprocess import PIPE
import sys
import time
from typing import (
    Any,
    Tuple,
)
import warnings

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy.optimize as opt
from scipy.stats import linregress
from scipy.integrate import quadrature as quad

import thunderboltz as tb
from thunderboltz.input import CrossSections
from thunderboltz import parameters
from thunderboltz.parameters import tb_parameters
from thunderboltz.parameters import wrap_parameters
from thunderboltz import parsing
from thunderboltz.parallel import MPRunner
from thunderboltz.parallel import visit
from thunderboltz.plotting.figure_gen import plot_timeseries
from thunderboltz.plotting.figure_gen import plot_vdfs

# Physical Constants
ME = 9.1093837e-31 # kg
QE = 1.60217663e-19 # C or J/eV
AMU_TO_KG = 1.6605e-27 # kg / amu

# Elastic angular distribution function names and parameters
EADF = {"He_Park": ("Park", [0.283, 0.667, 0.0307, 16.971, 6.59, 29.02, 0.0258, 0.295, 0.00328, 1.794]),
        "default": (None, None)}

# Electron impact ionization energy sharing distribution function names and parameters
EESD = {"default": (None, None),
        "equal": ("Equal", None),
        "uniform": ("Uniform", None)}

# Parameters specifically intended for cross section set construction
CS_PARAMS = ["n", "analytic_cs", "proper_sample", "nsamples", "egen",
             "mix_thresh", "eadf", "ECS", "fixed_background"]


class ThunderBoltz(MPRunner):
    """ThunderBoltz 0D DSMC simulation wrapper.

    Args:
        directory (str): The path to a directory that will host
            ThunderBoltz compiled, source, input, and output files.
        cs (CrossSections): The set of cross section information
            required for this simulation. Optionally supplied as an alternative
            to ``indeck``, default is an empty CrossSections object.
        out_file (str): Optional file base name for ThunderBoltz stdout buffer,
            default is ``"thunderboltz"``.
        monitor (bool): Runtime flag, when set to ``True`` an empty ``monitor``
            file will be generated in the simulation directory.
            Deleting this file will cause the ThunderBoltz process to exit,
            but allow the wrapper to continue execution. This is useful
            performing several simulation calculations sequentially,
            but manual exit is required for each one, or if post processing
            is required immediately after ThunderBoltz exits.
        live (bool): Run and update time series plotting GUI during simulation.
        ts_plot_params (list[str]): The default output parameters to be plotted
            by :func:`plot_timeseries`.
        **params:
            ThunderBoltz simulation parameters. Any attributes of
            :class:`~.thunderboltz.parameters.TBParameters` or
            :class:`~.thunderboltz.parameters.WrapParameters` can be passed here.
    """
    #: Name of the simulation output file produced by ThunderBoltz
    logfile = "thunderboltz.log"
    #: Files to be read when tabular data is requested
    output_files = ["thunderboltz.out", "Particle_Type", "Counts.dat", logfile]
    #: Time series plot parameters
    ts_plot_params = ["MEe", "mobN", "a_n"]

    def __init__(self, directory=None, cs=None, out_file=None,
                 monitor=None, live=None, live_rate=None, ts_plot_params=None, **params):
        """Initialize the ThunderBoltz object."""
        # Initialize data structures
        #: Particle-specific times series data
        self.particle_tables = []
        self.kinetic_table = None #: Banner output data
        self.timeseries = None #: All tick-by-tick simulation data
        self.vdfs = None #: Particle velocity dump data
        #: Particle velocity data intended for particle initialization
        self.vdf_init_data = None

        #: Time step at which steady-state calculations are considered converged
        self.time_conv = None
        self.counts = None #: pd.DataFrame: Table of collision counts
        # Initialize runtime logging params
        self.elapsed_time = None #: Elapsed wall-clock time of calculation
        self.runtime_start = None #: Date/Time of calculation start
        self.runtime_end = None #: Date/Time of calculation end
        # Initialize runtime params
        self.out_file = out_file #: Name for ThunderBoltz stdout file

        # Plotting objects and settings
        #: (bool): Run and update time series plotting GUI during simulation.
        self.live = live
        #: (bool): Run and update reaction rate plotting GUI during simulation.
        self.live_rate = live_rate
        #: The figure object for the time series plot.
        self.ts_fig = None
        #: The figure object for the rate plot
        self.rate_fig = None
        #: List of functions that are called every time banner output is
        #: updated.
        self.callbacks = []

        self.directory = directory #: Simulation directory
        #: Recorded thunderboltz warnings read in from output files
        self.err_stack = []
        if ts_plot_params: self.ts_plot_params = ts_plot_params

        # The Popen subprocess that will execute ThunderBoltz
        self._proc = None
        #: (bool): Option to create temp file during run that causes safe exit upon deletion
        self.monitor = monitor

        # Grab hyper param defaults
        all_params = copy.deepcopy(wrap_parameters)
        # Update with default thunderboltz params
        all_params.update(copy.deepcopy(tb_parameters))
        # Update with user initialized thunderboltz params
        all_params.update(copy.deepcopy(params))
        # Remember initial init parameters
        self.initial_params = copy.deepcopy(params)
        # Make empty CrossSections object
        self.cs = cs
        if cs is None:
            self.cs = CrossSections()
        # Set input for the first time
        self.hp = {}
        self.tb_params = {}
        self.set_(**all_params)

    def set_(self, **p):
        """Update parameters, call appropriate functions ensuring
        input parameters are self-consistent.

        Args:
            **p: Optional keyword parameters to update the calculator.
                can be any of :class:`~.thunderboltz.parameters.TBParameters` or
                :class:`~.thunderboltz.parameters.WrapParameters`.
        """
        # Set thunderboltz and hyper params appropriately
        self.hp.update({k: v for k, v in p.items()
            if k in wrap_parameters})
        self.tb_params.update({k: v for k, v in p.items()
            if k in tb_parameters})
        # For now, also allow arbitrary parameters for log files
        self.hp.update({k: v for k, v in p.items()
            if k not in tb_parameters})
        # Update directory
        if "directory" in p:
            self.directory = p["directory"]

        # Determine if special parameters changed
        resample = CS_PARAMS + ["indeck"]
        if any(i in resample for i in p):
            self._load_indeck()

        # Adjust differential models if necessary
        if ("eesd" in p or "eadf" in p) and len(self.cs.table):
            self._update_differential_models()

        # No matter what, input files should be considered unwritten
        # and input re-constrained
        self.written_input = False
        self._constrain_input()

        # Interpret velocity init file request (must occur after NP is set)
        if self.hp["vdf_init"]: self._prepare_vdf_init()

    def reset(self):
        """Reset output data for a new run."""
        self.particle_tables = []
        self.kinetic_table = None
        self.timeseries = None
        self.vdfs = None
        self.vdf_init_data = None
        self.time_conv = None
        self.counts = None
        self.elapsed_time = None
        self.runtime_start = None
        self.runtime_end = None
        self.ts_fig = None
        self.rate_fig = None
        self.callbacks = []
        self.err_stack = []
        self._proc = None

    def get_sim_param(self, key):
        """Return the value of a simulation parameter.

        Raises:
            IndexError: if the parameter is not set.
        """
        if key in self.hp:
            return self.hp[key]
        if key in self.tb_params:
            return self.tb_params[key]

        # If not found, raise error
        raise IndexError(f"'{key}' not a set simulation parameter.")

    def read_tb_params(self, fname, ignore=[]):
        """Takes file name of an input deck, updates the simulation
        parameters and returns the simulation parameters which were
        read from the file ``fname``.

        Args:
            fname (str): The name of the indeck file to read.
            ignore (list[str]): Don't read certain ThunderBoltz
                params, e.g. ``["MP", "QP"]`` would ignore the
                mass and charge parameters in an indeck file.
        Returns:
            dict: ``tb_params``.
        """
        # Read the input file
        with open(fname, "r") as f:
            lines = f.readlines()
        # Drop lines in ignore, comments, and the cross section info table
        # and form parameter key, value pairs in tuples. Keys may repeat at
        # this point
        keys = [l.split()[0] for l in lines if l.split()[0] not in ignore+["CC", "CS"]]
        vals = [l.split()[1:] for l in lines if l.split()[0] not in ignore+["CC", "CS"]]

        # Consolidate the values of repeated keys into a list of lists lines
        multi_keys = [k for k in keys if keys.count(k) > 1]
        multi_vals = [v for k, v in zip(keys, vals) if keys.count(k) > 1]
        tb_params = {k: [] for k in set(multi_keys)}
        for mk, mv in zip(multi_keys, multi_vals):
            tb_params[mk].append([infer_param_type(p) for p in mv])

        # Add the rest of the keys
        for k, v in zip(keys, vals):
            if k in multi_keys: continue
            tb_params[k] = [infer_param_type(p) for p in v]
            if len(v) == 1: tb_params[k] = tb_params[k][0]

        self.tb_params.update(tb_params)
        return tb_params

    def _prepare_vdf_init(self):
        """Infer type of vdf_init request and populate vdf_init_data
        accordingly."""
        # Create directory to store simulation veloctity init data.
        os.makedirs(pjoin(self.directory, "velocity_init"), exist_ok=True)
        # Convert to a list of each particle type
        if np.array(self.hp["vdf_init"], dtype=object).ndim < 2:
            self.hp["vdf_init"] = [self.hp["vdf_init"]]

        self.vdf_init_data = []
        self.tb_params["LV"] = []
        for vdf_info, ptype in self.hp["vdf_init"]:
            if isinstance(vdf_info, int):
                # Interpret as step
                if not vdf_info in self._check_available_vdfs():
                    raise RuntimeError("Requested vdf_init step not available")
                self.get_vdfs(steps=vdf_info, particle_type=ptype, sample_cap=None)
                self.vdf_init_data.append([
                    self.vdfs.iloc[self.vdfs.step == vdf_info, -3:].copy(),
                    ptype,
                ])
            elif isinstance(vdf_info, str):
                # Interpret as path to vdf file
                self.vdf_init_data.append([
                    parsing.extract_vdf(vdf_info, sample_cap=None),
                    ptype,
                ])
            elif isinstance(vdf_info, pd.DataFrame):
                self.vdf_init_data.append([vdf_info.reset_index(drop=True), ptype])
            elif isinstance(vdf_info, np.ndarray) or isinstance(vdf_info, list):
                self.vdf_init_data.append([
                    pd.DataFrame(vdf_info, columns=["vx", "vy", "vz"]),
                    ptype,
                ])

            if self.hp["downsample"]:
                # Truncate init data to match NP
                self.vdf_init_data[-1][0] = (
                    self.vdf_init_data[-1][0].iloc[:self.tb_params["NP"][ptype],:].copy()
                )

            # Add requests to tb input deck
            fname = pjoin("velocity_init", f"ParticleType{ptype}.dat")
            self.tb_params["LV"].append([fname, ptype])


    def _load_indeck(self):
        """Sample and data from indeck options if available."""
        indeck = self.hp["indeck"]
        if callable(indeck):
            # Auto generate CS and indeck with function `indeck`
            # and these parameters
            csps = {k: v for k, v in self.hp.items() if k in CS_PARAMS}
            self.cs, tbp = indeck(**csps)
            # Ignore user specified params
            tb_updates = {k: v for k, v in tbp.items() if k not in self.initial_params}
            self.tb_params.update(tb_updates)
        elif indeck:
            # Interpret as path to pre-written input data
            self.cs = CrossSections(input_path=indeck, read_cs_data=True)
            # Ignore init params set by user
            ignored = list(self.initial_params.keys())
            self.read_tb_params(
                    pjoin(indeck, self.cs.input_fname), ignore=ignored)
        # Otherwise, don't configure indeck data

    def _update_differential_models(self):
        """Update cross section process tables with parameters."""
        # Elastic
        name, params = EADF[self.hp["eadf"]]
        if isinstance(params, list):
            params = " ".join(str(p) for p in params)
        msk = self.cs.table.rtype.str.contains("Elastic")
        self.cs.table.loc[msk, ["model_name", "params"]] = (name, params)
        # Ionization
        name, params = EESD[self.hp["eesd"]]
        if isinstance(params, list):
            params = " ".join(str(p) for p in params)
        msk = self.cs.table.rtype.str.contains("Ionization")
        self.cs.table.loc[msk, ["model_name", "params"]] = (name, params)

    def _impose_nprod_crit(self):
        """Minimum requirement for the product of each species particle count."""
        # Aliases
        tb = self.tb_params
        hp = self.hp
        # Use some defaults if not defined
        if "DE" not in hp:
            hp["DE"] = 0.1 # eV
        if "EP_0" not in hp:
            hp["EP_0"] = 10.
        if "Nmin" not in hp: # Npairs multiplier (>=1)
            hp["CSTR_P"] = 1.0
        # Reduced mass (kg)
        m1, m2 = tb["MP"][0], tb["MP"][1]
        mu = m1*m2/(m1+m2) * AMU_TO_KG # amu to kg
        if not "L" in tb:
            raise RuntimeError("no L specified for DT calculation")

        cs_sets = self.cs.data
        cs_df = pd.DataFrame()
        # Collect all cross sections into one data frame.
        for key, df_ in cs_sets.items():
            df = df_.copy()
            e, cs = df.values.T
            df["rtype"] = key
            # Calculate vsig(eps) for each cs
            df["vsig"] = np.sqrt(2*e*QE/mu)*cs
            cs_df = pd.concat((cs_df, df), ignore_index=True)

        vsig_min_max = np.min(cs_df.groupby("rtype").vsig.agg("max"))
        vsig_sum_max = np.sum(cs_df.groupby("rtype").vsig.agg("max"))
        # Convert reduced field to Vm^2
        EoN = 1e-21*hp["Ered"]
        # Condition: only accelerate EP_0 electrons a max of DE each step
        Ne = hp["Nmin"]*np.sqrt(2*QE*hp["EP_0"]/ME)*2*QE*EoN/(QE*hp["DE"]*vsig_min_max)

        if "pct_ion" in hp:
            tb["NP"][1] = int(np.sqrt(Ne/hp["pct_ion"]))
            tb["NP"][0] = int(hp["pct_ion"]*np.sqrt(Ne/hp["pct_ion"]))

        # Compute E with NP and L
        n_gas = tb["NP"][hp["gas_index"]]/tb["L"]**3
        tb["E"] = EoN*n_gas
        # Compute DT with E, EP_0, DE
        tb["DT"] = QE*hp["DE"]/(QE*tb["E"]) * np.sqrt(ME/(2*QE*hp["EP_0"]))

        # Scale individual parameters beyond calibrated values.
        # Note, this will break E/N initial condition
        for k, v in self.hp["pc_scale"].items():
            if k == "NP":
                for i in range(len(tb[k])):
                    tb[k][i] = int(v*tb[k][i])
            else:
                tb[k] = v*tb[k]

    def _constrain_input(self):
        """Use alternative parameters to derive those required
        for the ThunderBoltz input files."""
        tb = self.tb_params
        hp = self.hp
        # Check gas index
        if hp["gas_index"] > 0 and len(tb["NP"]) == 1:
            # Assume there must be a gas for reduced calculations
            hp["gas_index"] = 0

        if self.hp["autostep"]:
            if not self.hp["DE"]:
                raise RuntimeError("Specify max DeltaE per time step for autostep")
            if not self.hp["Ered"]:
                raise RuntimeError("Specify reduced field (E/N) for autostep")
            if not self.hp["pct_ion"]:
                raise RuntimeError("Specify percent ionization (pct_ion) for autostep")
            # Calculate Nprod constraint and time step
            self._impose_nprod_crit()
        elif hp["pct_ion"] and hp["NN"]:
            # Calculate # of particles
            tb["NP"] = [int(hp["NN"]*hp["pct_ion"]), hp["NN"]]

        # Gas density
        n_gas = tb["NP"][hp["gas_index"]]/tb["L"]**3

        # Calculate other derived parameters, if need be
        if hp["duration"]:
            tb["NS"] = hp["duration"] // tb["DT"]

        if hp["Ered"] is not None:
            # Calculate E-field in V/m such that particle 0
            # is accelerated in the +z direction
            tb["E"] = int(np.sign(tb["QP"][0]))*1e-21*hp["Ered"]*n_gas

        if hp["Bred"] is not None:
            if not isinstance(hp["Bred"], list):
                raise RuntimeError("Please specify all three components of the "
                                   "reduced magnetic field.")
            # Calculate absolute B-field in Gauss (1 Gauss = 1e23 Hx m^-3)
            tb["B"] = [1e-27*Bred_i*n_gas for Bred_i in hp["Bred"]]

        # Round NP if its not already an integer
        tb["NP"] = [round(n) for n in tb["NP"]]

    def _validate_input(self):
        """Make sure the ThunderBoltz indeck will be valid before running."""
        # Alias
        tb = self.tb_params
        # Make sure QP and MP are correctly specified
        SP = len(tb["QP"])
        if not SP == len(tb["MP"]):
            raise RuntimeError("QP and MP arguments do not match in length")
        pass
        # All other multi-argument settings should not exceed these in length match SP parameter
        if any(len(tb[k]) > SP for k in ["NP", "TP"]):
            raise RuntimeError("Some multi-argument parameters do not match in length")

        # Infer number of species from number of charge and mass arguments
        tb["SP"] = SP

        # If multi-argument settings are shorter in length than inferred SP, set to 0
        arg_diff = SP - len(tb["NP"])
        if arg_diff:
            tb["NP"].extend([0]*arg_diff)
        arg_diff = SP - len(tb["TP"])
        if arg_diff:
            tb["TP"].extend([0.]*arg_diff)
        arg_diff = SP - len(tb["VV"])
        if arg_diff:
            tb["VV"].extend([0.]*arg_diff)

        # Ensure that the indices in the CS indeck portion do not exceed SP
        cs_sp_count = self.cs.table[["r1", "r2", "p1", "p2"]].max().max() + 1
        if cs_sp_count > SP:
            raise IndexError("Collision model species indices exceeds "
                             "specified number of species: "
                             f"({cs_sp_count} > {SP})")

    def set_fixed_tracking(self):
        """Change all reaction species indices of differing reactant values
        to be between only particle 0 and 1 (e.g. 0+1->0+2 is changed to 0+1->0+1)."""
        diff_msk = self.cs.table.r1 != self.cs.table.r2
        if len(diff_msk):
            self.cs.table.loc[diff_msk, ["r1", "r2", "p1", "p2"]] = [0, 1, 0, 1]

    def write_input(self, directory):
        """Write all the input files into a directory with
        the current settings.

        Args:
            directory (str): The path to the simulation directory.
        """
        # Write cross section files
        self.cs.write(directory=directory)
        # Write particle velocity init files if necessary.
        if self.vdf_init_data is not None:
            for i, (vdf_dat, ptype) in enumerate(self.vdf_init_data):
                fname = pjoin(directory, self.tb_params["LV"][i][0])
                vdf_dat.to_csv(fname, header=False, index=False)

        # Write simulation params into input.in
        with open(pjoin(directory, "input.in"), "w") as f:
            for k, v in self.tb_params.items():
                if v is None or v == [] or k not in tb_parameters:
                    continue
                if isinstance(v, list):
                    if isinstance(v[0], list):
                        for line in v:
                            args = " ".join(str(s) for s in line)
                            f.write(f"{k} {args}\n")
                    else:
                        v = " ".join(str(s) for s in  v)
                        f.write(f"{k} {v}\n")
                else:
                    f.write(f"{k} {v}\n")
            # Write CS arguments
            cs_deck = self.cs.get_deck()
            f.write(cs_deck)

        self.written_input = True

    def compile_from(self, src_path, debug=False):
        """Copy TB files from src_path and compile
        in simulation directory.

        Args:
            src_path (str): The location of the source files
                to compile from.
            debug (bool): If ``True``, compile C++ with the ``-g``
                debug flag.

        Raises:
            RuntimeError: if there is a compilation issue.
        """
        # Get source file names
        src_files = ls(src_path)
        # Copy each to simulation directory
        for sf in src_files:
            if ".cpp" in sf or ".h" in sf:
                src_file_path = pjoin(src_path, sf)
                dest_file_path = pjoin(self.directory, sf)
                shutil.copy(src_file_path, dest_file_path)

        # Go to simulation directory and compile
        pwd = os.getcwd()
        os.chdir(self.directory)

        # Compile (hard-coded for security reasons).
        cmd = "g++ -std=c++17 DSMC0D.cpp -o thunderboltz.bin -Wall -Werror -Wsign-compare"
        # Debug
        if debug:
            cmd = "g++ -g -std=c++17 DSMC0D.cpp -o thunderboltz.debug -Wall -Werror -Wsign-compare"
        cerr = os.system(cmd)
        if cerr:
            raise RuntimeError("Failed to compile.")

        os.chdir(pwd)

    def compile_debug(self):
        """Prepare all files and compile ThunderBoltz with -g debug flag."""
        self.run(dryrun=True, debug=True)

    def add_callback(self, f):
        """Add a function to the list of functions that will be
        called during banner output.

        Args:
            f (callable[,]): The function that will be called. It
                should accept no arguments and return no arguments.
        """
        self.callbacks.append(f)

    def run(self, src_path=None, bin_path=None, out_file="thunderboltz",
            monitor=False, dryrun=False, debug=False, std_banner=False,
            live=False, live_rate=False):
        """Execute with the current parameters in the simulation directory.
           The internal API ThunderBoltz version will be used in lieu of
           user-provided binary/source files.

        Args:
            src_path (str): Optional path to source files to copy into the
                simulation directory. The source is then compiled there.
            bin_path (str): Optional path to binary executable to copy
                into the simulation directory.
            out_file (str): The file name for stdout buffer of the
                ThunderBoltz process.
            monitor (bool): Runtime flag, when set to `True` will generate
                an empty `monitor` file in the simulation directory. Deleting this
                file will cause the ThunderBoltz process to exit, but allow
                the wrapper to continue execution. This is useful performing
                several simulation calculations sequentially, but a manual exit
                is required for each one.
            dryrun (bool): Setup all the files for the calculation, but do not
                run the calculation.
            debug: (bool): Compile with C++ ``-g`` debug flag.
            std_banner (bool): Toggle banner output streaming to stdout in
                addition to being written to the ``out_file`` buffer.
            live (bool): Run and update time series plotting GUI during simulation.
            live_rate (bool): Run and update rate plotting GUI during simulation.

        Raises:
            RuntimeError: if there is no simulation directory set or if
                the one provided does not exist.
        """
        if not self.directory:
            raise RuntimeError(
                "Simulation directory has not been established")
        elif not os.path.isdir(self.directory):
            raise RuntimeError(
                "Simulation directory does not exist")

        # Overwrite monitor, out_file, and live options
        if self.monitor is None:
            self.monitor = monitor
        if self.out_file is None:
            self.out_file = out_file
        if self.live is None:
            self.live = live
        if self.live_rate is None:
            self.live_rate = live_rate

        # Validate parameters and write indeck files
        self._validate_input()
        if not self.written_input:
            self.write_input(self.directory)

        # Determine binary TB file
        if not bin_path:
            if not src_path:
                # Use internal package src
                src_path = tb.src_path
            # Create binary file from source
            self.compile_from(src_path, debug=debug)
        else:
            # Copy binary file from user path
            bin_file = Path(bin_path).parts[-1]
            shutil.copy(bin_path, pjoin(self.directory, bin_file))

        # Record time and log params
        self.runtime_start = datetime.now().strftime("%d-%m-%y_%H:%M:%S")
        self._log()

        # Do not run ThunderBoltz for dry runs
        if dryrun: return

        ##### Run the program #####

        # Sub-process commands hard-coded for security reasons
        args = ["./thunderboltz.bin", "input.in"]

        # Create callbacks for communication between the
        # ThunderBoltz thread and this parent thread.

        if self.monitor:
            # Create temporary monitor file
            with open(pjoin(self.directory, "monitor"), "w") as f:
                f.write("") # Write empty file
            self.add_callback(self._monitor_callback)

        if self.live:
            # Live time series plotting
            self.add_callback(self._live_ts_callback)
        if self.live_rate:
            # Live rate coefficient plotting
            self.add_callback(self._live_rate_callback)

        # Start precision timer
        t1 = time.perf_counter()

        # Start the ThunderBoltz subprocess
        self._interactive_subprocess(args, tee=std_banner,
            std_buf=pjoin(self.directory, self.out_file+".out"),
            err_buf=pjoin(self.directory, self.out_file+".err"),
        )

        # After subprocess exits, record wall-clock time.
        self.elapsed_time = time.perf_counter() - t1
        self.runtime_end = datetime.now().strftime("%d-%m-%y_%H:%M:%S")

        # Update log files with elapsed times
        self._log()

    def _interactive_subprocess(self, args, tee=True, std_buf=None,
            err_buf=None):
        """Open a subprocess for the ThunderBoltz simulation and
        pipe the output back to the parent. Optionally allow for
        'tee'-like behavior and also print out process
        stdout at the same time."""
        # Start process in simulation directory
        with visit(self.directory):
            self._proc = subprocess.Popen(args, stdout=PIPE, stderr=PIPE)

        p = self._proc # Alias
        # Ensure file reading doesn't wait for subprocess to close when
        # sending updates
        os.set_blocking(p.stdout.fileno(), False)

        if std_buf: f = open(std_buf, "a", buffering=1)

        def run_callbacks():
            lines = p.stdout.readlines()
            txt = "".join(s.decode() for s in lines)
            if std_buf: f.write(txt)
            if self.callbacks: [c() for c in self.callbacks]
            if tee: print(txt, end="")

        while p.poll() is None:
            # Release lock to the subprocess if running
            # on one core.
            time.sleep(2)

            run_callbacks()

        # Run buffering one last time.
        run_callbacks()

        if std_buf: f.close()

        # Log errors into .err file after failure and raise
        # as warning in the parent
        err = "\n".join(s.decode() for s in p.stderr.readlines())
        if err_buf:
            with open(err_buf, "w") as e:
                e.write(err)
        elif err:
            warnings.warn(err, RuntimeWarning)

        return p

    def _monitor_callback(self):
        """Exit the ThunderBoltz process when the
        monitor file is removed by the user."""
        p = self._proc # Alias

        # Check if monitor file is still there, or if process ended.
        monfile = pjoin(self.directory, "monitor")
        if not os.path.isfile(monfile) or p.poll() is not None:
            # At this point, the process is either terminated, or should be
            if p.poll() is None:
                p.kill()
            # And the monitor file is either deleted, or should be
            if os.path.isfile(monfile):
                os.remove(monfile)

    def _live_ts_callback(self):
        """Update the time series GUI."""
        # Make sure output files have data
        if not os.path.isfile(
                pjoin(self.directory, self.out_file+".out")):
            return # Don't try to plot anything

        # Clear plots for data redraw if not initialized yet
        if self.ts_fig: [a.clear() for a in self.ts_fig.axes]
        self.plot_timeseries()
        if self.ts_fig: self.ts_fig.canvas.draw()
        plt.pause(0.2)

    def _live_rate_callback(self):
        """Update the rate plot GUI."""
        if not os.path.isfile(
                pjoin(self.directory, self.out_file+".out")):
            return # Don't try to plot anything

        # Clear plots for data redraw if not initialized yet
        if self.rate_fig: [a.clear() for a in self.rate_fig.axes]
        self.plot_rates()
        if self.rate_fig: self.rate_fig.canvas.draw()
        plt.pause(0.2)

    def read_stdout(self, fname):
        """Read the banner output data.

        Args:
            fname (str): The name of the ``.out`` file to read.

        Returns:
            :class:`pandas.DataFrame`: The banner data.

        Note:
            If ThunderBoltz warnings are found (e.g. particle overload),
            a message will be appended to ``err_stack``.
        """
        fpath = pjoin(self.directory, fname)
        with open(fpath, "r") as f:
            # Get to data lines
            l = f.readline()
            while "0, t" not in l:
                if "Too many particles!" in l:
                    self.err_stack.append("particle overload")
                if not l:
                    self.kinetic_table = pd.DataFrame()
                    return

                l = f.readline()
            # Read the rest of the lines
            lines = [l] + f.readlines()

        if not lines:
            self.kinetic_table = pd.DataFrame()
            return

        # Convert line data to DataFrame
        if "Too many particles!" in lines[-1]:
            lines = lines[:-1]
            self.err_stack.append("particle overload")

        lines = [l.replace(" ", "") for l in lines]
        # Ignore comments
        lines = [l for l in lines if not l.lstrip().startswith("#")]
        txt = "".join(lines)
        cmax = len(lines[0].split(","))
        a = np.genfromtxt(io.StringIO(txt), delimiter=",",
            usecols=range(0, cmax+1, 2), autostrip=True, ndmin=2)
        columns = ["step"] + [lines[0].split(",")[i] for i in range(1, cmax, 2)]
        self.kinetic_table = pd.DataFrame(a, columns=columns)
        return self.kinetic_table

    def read_particle_table(self, i):
        """Read species specific output data, including
        density, velocity, displacement, energy, and temperature.

        Args:
            i (int): The species index.

        Returns:
            :class:`pandas.DataFrame`: The particle data for
            species ``i``.
        """

        fpath = pjoin(self.directory, f"Particle_Type_{i}.dat")
        with open(fpath, "r") as f:
            column_names = f.readline().replace(",", "").split()
            lines = f.readlines()

        lines = [l.replace(" ", "") for l in lines]
        if not lines: return
        txt = "".join(lines)
        a = np.genfromtxt(io.StringIO(txt), delimiter=",",
                autostrip=True, ndmin=2)
        self.particle_tables[i] = pd.DataFrame(a, columns=column_names)
        return self.particle_tables[i]

    def read(self, directory=None, read_input=True, read_cs_data=False, only=None):
        """Read the simulation directory of a ThunderBoltz run, possibly
        all of its input and output files.

        Args:
            directory (str): The location of the simulation directory
                from which to read.
            read_input (bool): Whether or not to read any input data.
            read_cs_data (bool): Whether or not to read cross section data.
                This can be expensive, and often isn't necessary.
            only (list or None): Only read certain types of files. Default is
                ``["thunderboltz.out", "Particle_Type", "Counts.dat", ""thunderboltz.log"]``.
        """
        if directory: self.directory = directory

        if read_input:
            # Read cross sections input first
            self.cs.read(self.directory, read_cs_data=read_cs_data)
            self.read_tb_params(pjoin(self.directory, self.cs.input_fname))
        # Read other files
        logfiles = ls(self.directory)
        if only:
            logfiles = [o for o in only if o in logfiles]
        for filename in logfiles:
            fpath = pjoin(self.directory, filename)
            if filename == "Counts.dat":
                counts_dat = np.genfromtxt(fpath, delimiter=",")
                if counts_dat.ndim == 1:
                    counts_dat = counts_dat.reshape(1, -1)
                self.counts = pd.DataFrame(counts_dat,
                    columns =["step"]+self.cs.table.csfile.to_list())
            elif filename == "thunderboltz.out":
                self.read_stdout(filename)
            elif filename == "thunderboltz.log":
                self.read_log(filename)

        # Read particle type files in order
        if only and "Particle_Type" not in only:
            return
        pfiles = [f for f in ls(self.directory) if "Particle_Type" in f]
        self.particle_tables = [None]*len(pfiles)
        for i in range(len(pfiles)):
            self.read_particle_table(i)

    def get_particle_tables(self):
        """Return the particle table data for each species
        in a list.

        Returns:
            list[ :class:`pandas.DataFrame`]: The list of particle
            table data. Each table with have columns with keywords
            matching the attributes of
            :class:`~.thunderboltz.parameters.ParticleParameters`.
        """
        # Implicitly call the main output parsing code.
        self.get_timeseries()
        return self.particle_tables

    def get_timeseries(self):
        """Collect the relevant time series data from a ThunderBoltz simulation
        directory and add input parameter columns.

        Returns:
            :class:`pandas.DataFrame`: The table of time series data.
        """
        # Read output files
        self.read(only=self.output_files, read_input=False)
        # Make sure actual output files got read
        pt = self.particle_tables
        kt = self.kinetic_table

        if kt is None:
            raise RuntimeError("Missing output logs, make sure stdout and "
                               "particle tables are being written.")
        if pt == [] or pt[0] is None or (len(pt[0]) < 2 or len(kt) < 2):
            # Do not try to join until more data is available
            self.timeseries = pd.DataFrame()
            return self.timeseries


        # Create time series
        self.timeseries = pt[0].join(kt.drop("t", axis=1), how="inner")
        ts = self.timeseries

        # Add input and meta parameters
        self._tabulate(ts)

        # Compute some additional parameters
        DT = ts["t"].values[1] - ts["t"].values[0]
        # Calculate gas density
        ts["n_gas"] = self.tb_params["NP"][self.hp["gas_index"]]/ts.L**3
        # Calculate Reduced field
        ts.loc[:, "mobN"] = ts.Vzi.abs()/ts.E.abs()*ts.n_gas
        # Reduced Townshend ionization coefficient

        self._calculate_rates()

        ts["a_n"] = ts.k_i / ts.Vzi.abs()
        # Calculate bulk parameters if position data is available
        if "Rzi" in ts:
            zdrift = np.abs(np.concatenate(([0], np.diff(ts.Rzi)/DT)))
            ts["bulk_drift"] = zdrift
            ts["Rzi"] = ts.Rzi
            ts["mobN_bulk"] = zdrift/ts.E.abs()*ts.n_gas
            ts["a_n_bulk"] = ts.k_i / zdrift
        return ts

    def _calculate_rates(self):
        """Calculate rate coefficients from reaction counts."""
        ts = self.timeseries
        pt = self.particle_tables
        c = self.counts.set_index("step")

        # Find the shortest list (if files are still being written by process)
        npoints = min(len(ts), *[len(p) for p in pt], len(c))
        # Truncate counts to the shortest
        c = c.iloc[:npoints].copy()

        DT = ts["t"].values[1] - ts["t"].values[0]

        # Compute collision rate data
        dcdt = (c-c.shift())/DT
        ctab = self.cs.table
        # Get ionization collision files.
        ionz_files = ctab.loc[(ctab.rtype.str.contains("Ionization"))
                             |(ctab.csfile.str.contains("ion")), "csfile"]
        ionz_c = c[ionz_files].copy()
        # Compute total ionization count
        tot_ionz_c = ionz_c.sum(axis=1).values
        # Compute total interaction count over all processes
        total_c = c.sum(axis=1).values
        # For now use the most accurate time step values and assume constant DT
        ionz_rate = np.concatenate(([0], np.diff(tot_ionz_c)/DT))
        tot_rate = np.concatenate(([0], np.diff(total_c)/DT))

        # Generate specific rates for ionization and total processes
        k_scale = 1/(ts.n_gas*ts.Ni*ts.L**3)
        ts["k_i"] = ionz_rate*k_scale
        ts["k_tot"] = tot_rate*k_scale

        # Generate rates for all processes individually
        for i, row in ctab.iterrows():
            k_scaler = 1/(pt[row.r1].Ni * pt[row.r2].Ni * ts.L**3)[1:npoints]
            k_scaler[k_scaler == np.inf] = 0
            k_vals = (dcdt.iloc[1:npoints,i].values * k_scaler.values)
            ts.loc[1:len(k_vals),f"k_{i+1}"] = k_vals

    def get_ss_params(self, ss_func=None):
        """Get steady-state transport parameter values by averaging last section
        of time series. By default, the last fourth of the available data is
        considered to be steady-state. Standard deviations over this interval
        will be computed for each parameter in a new column with a "_std"
        suffix added to the column name.

        Args:
            ss_func (callable[:class:`pandas.DataFrame`, :class:`pandas.DataFrame`]):
                A function that takes in the numerical time series data and returns a
                truncated set of time series data that is considered to be at
                steady state.

        Returns:
            :class:`pandas.DataFrame`: The aggregated steady-state data for
            each output parameter along with columns specifying the input
            parameters.

        Raises:
            RuntimeWarning: if not enough steps are available to compute steady
                state statistics.

        Warning:
            Currently, the last quarter of the time series data is assumed to be
            in steady-state by default when calculating these steady-state parameters.
            One can verify that this is true by viewing the figures produced by
            :py:func:`plot_timeseries`. Otherwise, one may run the simulation for longer,
            or provide the appropriate criteria via ``ss_func``.
        """
        if self.timeseries is None:
            ts = self.get_timeseries()
            if not len(ts):
                # Return empty data
                return pd.DataFrame()

        ts = self.timeseries
        # Get numerical time series data
        tsn = ts.select_dtypes(include=[np.number]).reset_index(drop=True)
        containment = 0. # (False)
        endP = 30
        std_bar = 10
        if ss_func:
            ts_last = ss_func(tsn).copy()
        elif len(tsn) < 15:
            # take last fourth
            ts_last = tsn[tsn.index > 3*len(tsn)//4].copy()
        else:
            ts_last = tsn[tsn.index > 3*len(tsn)//4].copy()
            end_ts = ts_last.iloc[-endP:,:]
            es = end_ts.agg(["mean", "std"])
            mEm = es.loc["mean", "MEe"]
            mEs = es.loc["std", "MEe"]
            # Calculate a rough indicator of convergence (fractional %)
            contained = ts_last[(ts_last.MEe > mEm-std_bar*mEs) & (ts_last.MEe < mEm+std_bar*mEs)].copy()
            containment = len(contained)/len(ts_last)

        if len(ts_last) < 1:
            raise RuntimeError("Not enough steps to generate steady-state statistics")
        # Set time step after which stats are generated
        self.time_conv = ts_last.t.values[0]

        # Calculate time derivative based parameters
        LR = linregress(ts_last.t, ts_last.Rzi)
        zdrift = LR.slope
        anbf = "a_n_bulk_fit"
        mnbf = "mobN_bulk_fit"
        ts_last[anbf] = ts_last.k_i / zdrift
        ts_last[mnbf] = zdrift/ts_last.E.abs()*ts_last.n_gas
        # Update time series with these parameters
        ts.loc[ts.t >= self.time_conv, [anbf, mnbf]] = ts_last[[anbf, mnbf]]

        # Calculate steady-state statistics by aggregating over time
        stats = ts_last.agg(["mean", "std", lambda q: q.quantile(.9)])
        stats = stats.drop(["t", "step"], axis=1)
        stats = pd.concat((stats.loc["mean", :],
                           stats.loc["std", :].add_suffix("_std"),
        ), axis=0)

        # Correct uncertainty in fit with stderr
        stats[mnbf+"_std"] = stats[mnbf]*LR.stderr/zdrift
        stats[anbf+"_std"] = stats[anbf]*LR.stderr/zdrift

        # Save convergence criteria
        stats["containment"] = containment

        # Add input / meta parameters
        stats = pd.DataFrame([stats], index=[0])
        stats = self._tabulate(stats)
        return stats

    def get_etrans(self):
        """Return the energy weighted counts of each reaction computed
        for each time step.

        Returns:
            :class:`pandas.DataFrame`: A table of each process and the
            energy transfer through that channel as a proportion
            to the total energy transfer.
        """
        table = self.cs.table.set_index("csfile")
        weighted = self.counts.copy()
        # Loop through columns of reaction count table.
        for reaction in weighted.columns[1:]:
            weighted[reaction] *= table.loc[reaction, "B"]
        weighted.iloc[:, 1:] = weighted.iloc[:, 1:].divide(weighted.iloc[:, 1:].sum(axis=1), axis=0)
        return weighted

    def get_vdfs(self, steps="last", sample_cap=500000, particle_type=0, v=0):
        """Read the electron velocities arrays within a ThunderBoltz calculation.
        Velocity units are in m/s. If velocity dump files are found
        corresponding to ``steps``, update ``vdfs``.

        Args:
            steps (str, list[int], or int): Options for which time steps to
                read:

                * ``"last"``: Only read the VDF of the last time step
                * ``"first"``: Only read the VDF of the first time step
                * ``"all"``: Read a separate VDF for each time step.
                * ``list[int]``: Read VDF for each time step included in list.
                * ``int``: read VDF at one specific time step.

            sample_cap (int): Limit the number of samples read from the dump
                file for very large files. Default is 500000. If bool(sample_cap)
                evaluates to ``False``, then no cap will be imposed.
            particle_type (str, list[int], or int): Specify which kinds of species data
                should be read from.

                * ``int``: The particle type to read. Default is ``0``.
                * ``list[int]``: A set of particle types to read.
                * ``"all"``: Read all particle types.

            v (int): Verbosity -- 0: silent, 1: print file paths before reading.

        Returns:
            :class:`pandas.DataFrame`: The particle velocity dump data.

        Warning:
            Large files are truncated to the first ``sample_cap`` lines.
            It is assumed that the particle ordering in the dump files
            is not correlated with any velocity statistics, but
            this may not be the case when ``egen`` is on. In that case, ensure
            the entire velocity dump file is being read.
        """
        # Check what's available out of the requested
        available = self._check_available_vdfs(steps)
        if not len(available):
            return # Nothing to plot
        if self.vdfs is None:
            # Initialize empty DataFrame with named columns
            self.vdfs = pd.DataFrame({k: pd.Series([], dtype=dtype)
                for k, dtype in zip(["ptype", "step", "vx", "vy", "vz"],
                                    [np.int64]*2 + [np.float64]*3)})
                # Keep "vx ... vz" as the last three columns

        # Find which steps are missing from current vdf table
        unique = self.vdfs[self.vdfs.ptype==0].step.unique()
        missing = [step for step in available
                   if step not in unique]
        # Extract them from the simulation directory.
        missing_vdfs = parsing.extract_vdfs(
            self.directory, steps=missing, sample_cap=sample_cap,
            particle_type=particle_type, v=v,
        )

        # Only read the missing ones
        self.vdfs = pd.concat((self.vdfs, missing_vdfs), ignore_index=True)
        return self.vdfs

    def get_edfs(self, steps="last", sample_cap=500000):
        """Read the electron velocity distribution functions and return
        the component and total energy distributions within a
        ThunderBoltz calculation. Energy units are in eV. Invokes
        :func:`get_vdfs`.

        Args:
            steps (str, list[int], or int): Options for which time steps to
                read:

                * ``"last"``: Only read the VDF of the last time step
                * ``"first"``: Only read the VDF of the first time step
                * ``"all"``: Read a separate VDF for each time step.
                * ``list[int]``: Read VDF for each time step included in list.
                * ``int``: read VDF at one specific time step.

            sample_cap (int): Limit the number of samples read from the dump
                file for very large files. Default is 500000. If bool(sample_cap)
                evaluates to ``False``, then no cap will be imposed.

        Returns:
            :class:`pandas.DataFrame`: A table with the signed and unsigned
                energy components of each particle.
        """

        self.get_vdfs(steps, sample_cap)
        edf = self.vdfs.copy().rename(
            columns={k: k.replace("v", "E") for k in self.vdfs.columns[1:]})
        # Transform columns to eV, first signed energies
        m = self.tb_params["MP"][0]*AMU_TO_KG
        for col in ["Ex", "Ey", "Ez"]:
            edf[col+"_signed"] = 1/2 * m/QE * edf[col] * edf[col].map(np.abs)
        # Then the unsigned
        for col in ["Ex", "Ey", "Ez"]:
            edf[col] = edf[col+"_signed"].abs()
        # Total energy is the sum of its orthogonal components.
        edf["E"] = edf.Ex + edf.Ey + edf.Ez
        return edf

    def describe_dist(self, steps="last", sample_cap=500000):
        """Generate percentile and count statistics of the electron
        velocity / energy distribution for various time steps.

        Args:
            steps (str, list[int], or int): Options for which time steps to
                read:

                * ``"last"``: Only read the VDF of the last time step
                * ``"first"``: Only read the VDF of the first time step
                * ``"all"``: Read a separate VDF for each time step.
                * ``list[int]``: Read VDF for each time step included in list.
                * ``int``: read VDF at one specific time step.

            sample_cap (int): Limit the number of samples read from the dump
                file for very large files. Default is 500000. If bool(sample_cap)
                evaluates to ``False``, then no cap will be imposed.

        Returns:
            :class:`pandas.DataFrame`: A table with statistical descriptions of
            the velocity and energy distributions.
        """

        # Get energy distribution, this will implicitly update self.vdfs
        efs = self.get_edfs(steps, sample_cap)
        # Concatenate the two frames
        joined = pd.concat((self.vdfs, efs.iloc[:,-4:]), axis=1)
        stats = (joined
            .groupby(["ptype", "step"])
            .describe(percentiles=[0.01, .1, .25, .5, .75, .9, .99])
            .stack()
        )
        return stats

    def set_ts_plot_params(self, params):
        """Set the default series plotted by
        :meth:`~thunderboltz.ThunderBoltz.plot_timeseries`"""
        self.ts_plot_params = params

    def plot_timeseries(self, series=None, save=None, stamp=[], v=0,
            update=True):
        """Create a diagnostic plot of ThunderBoltz time series
        data.

        Args:
            series (list[str]):
                The y-parameters to plot onto the time series figure.
            save (str): Option to save the plot to a file path.
            stamp (list[str]): Option to stamp the figure with the value of
                descriptive parameters, e.g. the field, or initial
                number of particles. See :class:`~.thunderboltz.parameters.TBParameters`
                and :class:`~.thunderboltz.parameters.WrapParameters`.
            v (int): Verbosity -- 0: silent, 1: print file paths before
                plotting.
            update (bool): If set to ``False``, assume required data has
                already been parsed into ThunderBoltz frames.

        Returns:
            :class:`matplotlib.figure.Figure`: The timeseries figure object.
        """
        if series: self.ts_plot_params = series
        # If no data has been read yet, or program still running, read it
        process_running = self._proc and self._proc.poll() is None
        if update and (self.timeseries is None or process_running):
            self.get_timeseries()

        if not len(self.timeseries):
            return # No data

        if self.timeseries is None:
            raise RuntimeError("Timeseries data not read.")
        ts = self.timeseries.copy() # Alias

        # Convert mobility values
        ts["mobN"] = 1e-24*ts.mobN
        if "mobN_bulk" in ts:
            ts["mobN_bulk"] = 1e-24*ts.mobN_bulk
        if "mobN_bulk_fit" in ts:
            ts["mobN_bulk_fit"] = 1e-24*ts.mobN_bulk_fit

        # Create figure and ax objects, if necessary
        if not self.ts_fig:
            self.ts_fig, _ = plt.subplots(figsize=(14,9))
        fig = self.ts_fig # Alias

        # Plot timeseries with this figure
        plot_timeseries(fig, ts, series=self.ts_plot_params, save=save)
        # super title
        stamps = copy.deepcopy(self.hp)
        stamps.update(self.tb_params)
        if "directory" not in stamps:
            stamps["directory"] = Path(self.directory).parts[-1]
        stamps = {k: v for k, v in stamps.items() if k in stamp}
        # Use pandas Series to create a pretty string.
        stt = "\n".join(str(pd.Series(stamps, dtype="object")).split("\n")[:-1])
        fig.suptitle(stt, fontsize=10, fontproperties={"family": "monospace"})

        if save:
            # Save with the name of the simulation directory
            sname = Path(self.directory).parts[-1]
            if not os.path.isdir(save):
                os.makedirs(save)
            saveas = pjoin(save, sname+".pdf")
            if v: print(f"Saving time series in {saveas}")
            fig.savefig(saveas)
            plt.close(fig)

        return fig

    def plot_rates(self, save=None, stamp=None, v=0, update=True):
        """Create a diagnostic plot of ThunderBoltz time series
        data.

        Args:
            series (list[str]):
                The y-parameters to plot onto the time series figure.
            save (str): Option to save the plot to a file path.
            stamp (list[str]): Option to stamp the figure with the value of
                descriptive parameters, e.g. the field, or initial
                number of particles. See :class:`~.thunderboltz.parameters.TBParameters`
                and :class:`~.thunderboltz.parameters.WrapParameters`.
            v (int): Verbosity -- 0: silent, 1: print file paths before
                plotting.
            update (bool): If set to ``False``, assume required data has
                already been parsed into ThunderBoltz frames.

        Returns:
            :class:`matplotlib.figure.Figure`: The plot_rate figure object.
        """
        if self.rate_fig is None:
            self.rate_fig, ax = plt.subplots(figsize=(14, 10))

        ax = self.rate_fig.axes[0]

        if update: self.get_timeseries()
        ts = self.timeseries

        kkeys = [k for k in ts.columns if k.startswith("k_") and "i"
                 not in k]
        for col in ts.columns:
            if not col.startswith("k_") or "i" in col: continue

            ax.plot(ts.t, ts[col], label=col)
            ax.legend(fontsize=10)
            ax.set_yscale("log")
            ax.set_ylabel("Rate Coefficient (m$^3$/s)")
            ax.set_xlabel("Time (s)")

        if save:
            # Save with the name of the simulation directory
            sname = Path(self.directory).parts[-1]
            if not os.path.isdir(save):
                os.makedirs(save)
            saveas = pjoin(save, sname+"_rates.pdf")
            if v: print(f"Saving rate constants in {saveas}")
            self.rate_fig.savefig(saveas)
            plt.close(self.rate_fig)

        return self.rate_fig


    def plot_edf_comps(self, steps="last", sample_cap=500000,
                       bins=100, maxwellian=True, save=None):
        """Plot the directional components of the energy distribution function.

        Args:
            steps (str, list[int], or int): Options for which time steps to
                read:

                * ``"last"``: Only read the VDF of the last time step
                * ``"first"``: Only read the VDF of the first time step
                * ``"all"``: Read a separate VDF for each time step.
                * ``list[int]``: Read VDF for each time step included in list.
                * ``int``: read VDF at one specific time step.

            sample_cap (int): Limit the number of samples read from the dump
                file for very large files. Default is 500000. If bool(sample_cap)
                evaluates to ``False``, then no cap will be imposed.
            bins (int): Total number of bins to divide the energy space into.
            maxwellian (bool): Option to draw a maxwellian distribution
                with the same temperature for comparison.
            save (str): Optional location of directory to save the figure in.
        """
        # TODO: stamp with temperature for species.
        # TODO: Add variable-width bins
        # TODO: Add symmetrical log plots.
        # TODO: Add option to choose between the above.

        # XXX: For now, it is assumed that only VDFs for electrons exist.

        edfs = self.get_edfs(steps="last", sample_cap=500000)
        # Save pretty component names
        xlabels = [r"$\e" + f"psilon{c}$ (eV)" for c in ["_x", "_y", "_z", r"_{\rm tot}"]]
        figs = []
        steps = []
        for step, edf in edfs.groupby("step"):
            steps.append(step)
            fig, axs = plt.subplots(2, 2)
            steps.append(step)
            figs.append(fig)
            for i, (comp, ax) in enumerate(
                    zip(["Ex", "Ey", "Ez", "E"], axs.flatten())):
                # Get component series
                edf_c = edf[comp+"_signed"] if comp != "E" else edf["E"]
                # Plot the histogram
                edf_c.plot.hist(ax=ax, bins=bins)
                mean_e = edf["E"].mean()
                Ne = len(edf) # Counts of particles

                # Create maxwellian plot
                _, divs = np.histogram(edf_c.values, bins=bins)
                dE = divs[1] - divs[0]
                centers = divs[:-1] + dE/2
                prefactor=1
                # Mass of this species
                m = self.tb_params["MP"][0]
                if comp == "E":
                    # The speed distribution in units of energy
                    maxwellian = (lambda e: Ne*dE * e/(3/2*mean_e)**2 * np.exp(-abs(e)/(3/2*mean_e)))
                else:
                    # The 1D Gaussian velocity distribution in units of energy
                    maxwellian = lambda e: Ne*dE / (3*mean_e) * np.exp(-2/3*abs(e)/mean_e)
                maxw_y = np.vectorize(maxwellian)(centers)
                ax.plot(centers, maxw_y, label="Maxwellian")

                # View on a log scale
                ax.set_yscale("log")
                ax.set_ylabel(None)
                ax.set_xlabel(xlabels[i])
                ax.set_ylim(.9, None)
            fig.suptitle(f"step {step}", fontsize=13)
            fig.subplots_adjust(hspace=.18, wspace=.15)
            # Label axes
            axs[0,0].set_ylabel("Count")
            axs[1,0].set_ylabel("Count")
            fig.tight_layout()

            if save:
                sname = Path(self.directory).parts[-1]
                if not os.path.isdir(save):
                    os.makedirs(save)
                saveas = pjoin(save, sname+f"{step}_edf.pdf")
                print(f"Saving edf plots in {saveas}")
                sys.stdout.flush()
                fig.savefig(saveas)
                plt.close(fig)

        return figs, steps


    def plot_edfs(self, steps="last", sample_cap=500000,
            bins=100, plot_cs=False, save=None):
        """Plot the electron total energy distribution function, optionally
        include the provided cross sections for comparison.

        Args:
            steps (str, list[int], or int): Options for which time steps to
                read:

                * ``"last"``: Only read the VDF of the last time step
                * ``"first"``: Only read the VDF of the first time step
                * ``"all"``: Read a separate VDF for each time step.
                * ``list[int]``: Read VDF for each time step included in list.
                * ``int``: read VDF at one specific time step.

            sample_cap (int): Limit the number of samples read from the dump
                file for very large files. Default is 500000. If bool(sample_cap)
                evaluates to ``False``, then no cap will be imposed.
            bins (int): Total number of bins to divide the energy space into.
            maxwellian (bool): Option to draw a maxwellian distribution
                with the same temperature for comparison.
            save (bool): Optional location of directory to save the figure in.

        Returns:
            (Tuple[list[matplotlib.figure.Figure], list[int]): The list of
            figures and a list of their corresponding step indices.

        Note:
            It currently assumed that only data for one particle type is to be
            plotted.
        """
        edfs = self.get_edfs(steps, sample_cap)

        fig, ax = plt.subplots(figsize=(8,7))

        if plot_cs:
            # Plot it on the right y-axes
            self.cs.plot_cs(ax=ax.twinx())

        for step, df in edfs.groupby("step"):
            vals, divs = np.histogram(df.E, bins=bins)
            de = divs[1] - divs[0]
            evals = divs[:-1] + de/2
            ax.plot(evals, vals/de, label=f"step {step}")

        ax.legend()
        ax.set_ylabel("Particle number density (eV$^{-1}$)")
        ax.set_xlabel("Energy (eV)")

        fig.tight_layout()


    def plot_vdfs(self, steps="last", save=None, bins=100,
            sample_cap=500000):
        """Plot the joint distribution heat map between the x-y and x-z
        velocities.

        Args:
            steps (str, list[int], or int): Options for which time steps to
                read:

                * ``"last"``: Only read the VDF of the last time step
                * ``"first"``: Only read the VDF of the first time step
                * ``"all"``: Read a separate VDF for each time step.
                * ``list[int]``: Read VDF for each time step included in list.
                * ``int``: read VDF at one specific time step.

            sample_cap (int): Limit the number of samples read from the dump
                file for very large files. Default is 500000. If bool(sample_cap)
                evaluates to ``False``, then no cap will be imposed.
            bins (int): Total number of bins to divide the energy space into.
            save (str): Optional location of directory to save the figure in.

        Returns:
            (Tuple[list[matplotlib.figure.Figure], list[int]): The list of
            figures and a list of their corresponding step indices.
        """
        # Check whats available out of the requested
        self.get_vdfs(steps, sample_cap)
        available = self._check_available_vdfs(steps)
        # Confine plots to only the requested
        vdfs_data = self.vdfs[self.vdfs.step.isin(available)].copy()

        # Loop through included vdf groups
        figs = []
        steps = []
        for step, vdf in vdfs_data.groupby("step"):
            fig_xy = plot_vdfs(self.vdfs, bins=bins)
            fig_xz = plot_vdfs(self.vdfs, pair=["vx", "vz"], bins=bins)
            if save:
                sname = Path(self.directory).parts[-1]
                if not os.path.isdir(save):
                    os.makedirs(save)
                saveas = pjoin(save, sname+f"_xy{step}.pdf")
                print(f"Saving VDF joint plot in {saveas}", flush=True)
                fig_xy.savefig(saveas)
                plt.close(fig_xy)
                saveas = pjoin(save, sname+f"_xz{step}.pdf")
                print(f"Saving VDF joint plot in {saveas}", flush=True)
                fig_xz.savefig(saveas)
                plt.close(fig_xz)
            figs.extend([fig_xy, fig_xz])
            steps.append(step)
        return figs, steps

    def plot_cs(self, ax=None, legend=True, vsig=False,
                thresholds=False, save=None, **plot_args):
        r"""Plot the cross sections models.

        Args:
            ax (:class:`~.matplotlib.axes.Axes` or None): Optional
                axes object to plot on top of, default is None.
                If ax is None, then a new figure and ax object
                will be created.
            legend (bool): Activate axes legend if true,
                default is True.
            vsig (bool): Plot :math:`\sqrt{\frac{2\epsilon}{m_{\rm e}}}\sigma(\epsilon)`
                rather than :math:`\sigma(\epsilon)`.
            thresholds (bool): if True, plot energy units in thresholds.
            save (str): Optional location of the directory to save the plot in.
            **plot_args: Optional arguments passed to Axes.plot().
        Returns:
            :class:`matplotlib.axes.Axes`: The axes object of the plot.
        """
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.get_figure()
        ax = self.cs.plot_cs(
            ax, legend, vsig, thresholds, **plot_args)

        if save:
            sname = Path(self.directory).parts[-1]
            if not os.path.isdir(save):
                os.makedirs(save)
            saveas = pjoin(save, sname+"_cs.pdf")
            print(f"Saving cross section plot in {saveas}", flush=True)
            fig.savefig(saveas)
            plt.close(fig)

        ax.grid(which='major', color='gray', linestyle='-', alpha=0.5)
        ax.grid(which='minor', color='gray', linestyle='-', alpha=0.3)

        return ax

    def _check_available_vdfs(self, steps="all", get_types=False):
        """Check what steps are available base on required criteria."""
        available = []
        # First, check directory for file data
        try:
            for efile in ls(self.directory):
                if "Dump" in efile:
                    if "_0_" not in efile: continue
                    available.append(int(efile.split("_")[-1].split(".")[0]))
        except FileNotFoundError:
            pass
        # Then check for already parsed data
        if self.vdfs is not None:
            p0_only = self.vdfs.ptype == 0
            available = sorted(list(set(
                available + self.vdfs[p0_only].step.unique().tolist()
            )))

        if not len(available) or steps == "all":
            return available # No data available

        # Choose which time steps to include based on `steps`
        if steps == "last":
            return [max(available)]
        elif steps == "first":
            return [min(available)]
        elif isinstance(steps, list):
            return [s for s in steps if s in available]
        elif isinstance(steps, int):
            return [s for s in [steps] if s in available]

    def _tabulate(self, df):
        """Add params to DataFrame."""
        # Log inputs into table format
        for p, v in self.tb_params.items():
            if isinstance(v, list):
                df.loc[:, p] = " ".join(str(t) for t in v)
            else:
                df.loc[:, p] = v
        # Log meta params into table format
        for p, v in self.hp.items():
            if isinstance(v, list):
                df.loc[:, p] = " ".join(str(t) for t in v)
            elif isinstance(v, dict):
                df.loc[:, p] = str(v)
            elif callable(v):
                df.loc[:, p] = v.__name__
            else:
                df.loc[:, p] = v
        return df

    def _log(self, directory=None):
        """Write down the hyper parameters into a file."""
        if not directory:
            directory = self.directory
        with open(pjoin(directory, self.logfile), "w") as f:
            dat = {"hp": self.hp, "tb": self.tb_params, "elapsed_time": self.elapsed_time,
                    "runtime_start": self.runtime_start, "runtime_end": self.runtime_end}
            a = copy.deepcopy(self.hp["indeck"])
            # Convert functions to function names
            if callable(a):
                dat["hp"]["indeck"] = a.__name__
            # Convert any numpy/pandas datatypes into python primitives
            to_primitive(dat)
            # Serialize into log file
            json.dump(dat, f)
            self.hp["indeck"] = a

    def to_pickleable(self):
        """Return a picklable version of this object."""
        # It is currently entirely picklable
        return self

    def get_directory(self):
        """Return the path of the current simulation."""
        return self.directory

    def read_log(self, logfile):
        """Read json file from simulation directory. Update
        the corresponding settings in the ThunderBoltz object.

        Args:
            logfile (str): The path name of the logfile to read.
        """
        lpath = pjoin(self.directory, logfile)
        if os.path.isfile(lpath):
            with open(lpath, "r") as f:
                dat = json.load(f)
                self.hp.update(dat["hp"]) # Update in case params added
                self.tb_params.update(dat["tb"])
                self.elapsed_time = dat["elapsed_time"]
                self.runtime_start = dat["runtime_start"]
                self.runtime_end = dat["runtime_end"]

def read(directory, read_cs_data=False):
    """Create a ThunderBoltz object by reading from a
    ThunderBoltz simulation directory.

    Args:
        directory (str): The directory from which to
            initialize the ThunderBoltz object.
        read_cs_data (bool):
            When set to true, the reader will look for cs_data,
            default is ``False``

    Returns:
        :class:`~.thunderboltz.ThunderBoltz`: The ThunderBoltz object with
        tabulated data if available.
    """
    tb = ThunderBoltz()
    tb.read(directory, read_cs_data=read_cs_data)
    return tb

def query_tree(directory, name_req=None, param_req=None,
              read_cs_data=False, callback=None, agg=True):
    r"""Walk a directory tree and search for ThunderBoltz
    simulation directories to read. Either return a list of
    :class:`~.thunderboltz.ThunderBoltz` objects, or a custom aggregation of
    the output data.

    Args:
        directory (str): The root path to search for ThunderBoltz
            data in.

        name_req (callable[str,bool]): A requirement on the file path
            names to be included in the query. The callable accepts
            the file path of a thunderboltz simulation directory and
            should return ``True`` if that directory is to be included
            in the query.

            e.g. ``name_req=lambda s: "test_type_1" in s``
            would return only data in a subfolder ``test_type_1``.

        param_req (dict): A requirement on the
            parameter settings of the ThunderBoltz calculations.
            The dictionary corresponding to simulation parameters
            that must be set by the read ThunderBoltz object.

            e.g. ``param_req={"Ered": 100, "L": 1e-6}`` would only
            return data from calculations with a reduced field of
            100 Td and a cell length of 1 :math:`\mu{\rm m}`.

        callback: (callable[:class:`ThunderBoltz`, Any]):
            A function that accepts a ThunderBoltz object and
            returns the desired data.

        agg: If ``callback`` is set, attempt to aggregate the data
            based on the data type:

            ===============================   ============================
            callback Return Type              Behavior
            ===============================   ============================
            :class:`pandas.DataFrame`         Frames will be
                                              concatenated row-wise and
                                              one larger DataFrame will be
                                              returned.

            list[:class:`pandas.DataFrame`]   A list of frames the same
                                              length of the return value
                                              will be returned. The frame
                                              at index ``i`` will contain
                                              the concatenated data
                                              from each simulation returned
                                              by ``callable(tb)[i]``.

            list[Any]                         A list of lists will be
                                              returned. The list at index
                                              ``i`` will contain a list
                                              of items returned from each
                                              call to ``callable(tb)[i]``.

            Any                               Return values will be
                                              returned in a list.
            ===============================   ============================

            If ``agg`` is set to False, always return a list of callback data
            without any concatenation.

    Returns:
        list[:class:`ThunderBoltz`], or :class:`pandas.DataFrame`, or list[:class:`pandas.DataFrame`], or list[list[Any]]:
        See ``agg`` option for behavior. Default return type
        is list[:class:`ThunderBoltz`].

    """

    tbs = [] # For return type 1
    df = pd.DataFrame() # For return type 2
    dfs = [] # For return type 3
    tb_dat = [] # For return type 4
    # Default return type is list of TB objects
    rtype = tbs
    for path, dirs, files in os.walk(directory):
        # Ensure this is a ThunderBoltz directory
        if "cross_sections" in dirs and "thunderboltz.out" in files:
            # Ensure file path is matches requirements
            if name_req:
                if not name_req(path): continue

            # Read in data for this ThunderBoltz obj.
            tb = read(path, read_cs_data=read_cs_data)
            if param_req:
                valid_params = True
                for k, v in param_req.items():
                    set_val = tb.get_sim_param(k)
                    valid_params = (set_val == v) and valid_params
                if not valid_params: continue

            # Append object once determined to be included
            tbs.append(tb)

            # If desired, extract data from ThunderBoltz objects.
            if not callback:
                continue
            dat = callback(tb)
            if isinstance(dat, pd.DataFrame) and agg:
                # Build one big data frame with all the data
                df = pd.concat((df, dat), ignore_index=True)
                rtype = df
            elif isinstance(dat, (list, tuple)):
                # Build a list of either lists or DataFrames
                n = len(dat)
                is_df_list = all(isinstance(item, pd.DataFrame) for item in dat)

                # Initialize if necessary
                if not len(dfs): dfs = n*[pd.DataFrame()]
                if not len(tb_dat): tb_dat = [[] for _ in range(n)]
                for i, item in enumerate(dat):
                    if is_df_list and agg:
                        dfs[i] = pd.concat((dfs[i], item),
                            ignore_index=True)
                        rtype = dfs
                    else:
                        tb_dat[i].append(item)
                        rtype = tb_dat
            # Otherwise just return the list of ThunderBoltz objects

    # Return the correct datatype
    return rtype

def to_primitive(d):
    """Recursively convert collections with non primitive types
    back to python natives."""
    if isinstance(d, dict):
        iterable = d.items()
    elif isinstance(d, list):
        iterable = enumerate(d)

    for k, v in iterable:
        if isinstance(v, dict) or isinstance(v, list):
            # Recurse mutable collections in place
            to_primitive(v)
        elif isinstance(v, np.ndarray):
            d[k] = v.tolist()
        elif isinstance(v, np.number):
            d[k] = infer_param_type(v)
        elif isinstance(v, pd.DataFrame):
            d[k] = v.to_numpy().tolist()

def infer_param_type(x):
    """Sequential type checking for parameter io."""
    types = [int, float, str]
    for t in types:
        try:
            return t(x)
        except:
            continue
    raise RuntimeError("No type satisfied by {x}")




"""Main class for kinetic processes and their cross sections."""
import copy
import warnings

import numpy as np
from numpy import vectorize as vec
import pandas as pd


class Process(object):
    r"""A reaction process determined by reaction and product indices,
    a process type, a potential threshold, and a corresponding cross
    section specification.

    process_type: str
        Elastic | Inelastic | Ionization
    r1,r2,p1,p2: int (0,1,0,1)
        The indices of the reactants and products.
    threshold: float
        The threshold value for the process (e.g. the binding
        energy of an ionization process).
    cs_func: callable
        Function that returns the cross section for this process in
        :math:`\mathrm{m}^2` given an incident electron energy in
        eV (center of mass frame).
    cs_data: 2-D Array-Like
        Tabular cross section data with columns of energy (eV) and cross
        section (:math:`\mathrm{m}^2`).
    """
    # sample grid default range (orders of magnitude)
    SAMPLE_MIN = -4
    SAMPLE_MAX = 6

    def __init__(self, process_type, r1=0, r2=1, p1=0, p2=1,
                 threshold=0.0, cs_func=None, cs_data=None,
                 name=None, differential_process=None,
                 nsamples=250):
        self.process_type = process_type
        self.r1, self.r2, self.p1, self.p2 = r1, r2, p1, p2
        self.threshold = threshold
        self.differential_process = differential_process
        self.differential_parameters = None
        self.name = name
        self._cs_func = cs_func
        self._nsamples = nsamples
        if callable(cs_func):
            self.auto_sample()
        if cs_data is not None:
            self.data = self.to_cs_frame(cs_data)
            # Maintain threshold requirements
            self.zero_below_thresh()

        if cs_data is not None and cs_func:
            msg = "Both tabulated data and analytic functions were provided "
            msg += f"for process{self.name}"
            msg += "Tabulated data will be used"
            raise warnings.warn(msg)

        if "Elastic" in process_type and threshold > 0:
            raise ValueError("Process with a threshold > 0 cannot be elastic.")

    def to_df(self):
        """Convert to properly formatted pandas DataFrame.

        Returns:
            (:class:`pandas.DataFrame`): The process information.
        Raises:
            RuntimeError: if no data is available to produce a DataFrame
        """
        cols = ["csfile", "r1", "r2", "rtype", "B", "p1", "p2", "model_name", "params"]
        vals = [self.name+".dat", self.r1, self.r2, self.process_type,
                self.threshold, self.p1, self.p2, self.differential_process,
                self.differential_parameters]
        # Make sure cs_data is set
        if self.data is None:
            raise RuntimeError("Cross section data not provided")
        return pd.DataFrame([vals], columns=cols)

    def add_differential_parameters(self, name, params):
        """Typically differential processes require analytic forms
        due to the difficulty of extrapolation in several dimensions.
        Add free parameters into an analytic differential model here.

        Args:
            name: The name of the model for this differential process.
            params: The free parameters required for this differential model.
        """
        self.differential_process = name
        self.differential_parameters = params

    def sample_cs(self, e_points=None, grid_type="log dense", nsamples=None):
        """Sample self.cs_func on a grid of energies.

        Args:
            e_points (ArrayLike):
                Explicit energy (eV) grid points on which to sample.
            grid_type (str):

                =============== =============================================
                ``"log dense"`` sample ``nsamples`` near threshold up to 1MeV.
                   ``"simple"`` sample 0 eV - threshold - 1 MeV
                       ``None`` Behavior will automatically be determined
                                by ``auto_sample``.
                =============== =============================================

            nsamples (int):
                Override self.nsamples for this sampling call.
        """
        # Make sure there is a function to sample from.
        if not callable(self.cs_func):
            raise RuntimeError(
                "Cross section function is not callable/specified.")

        if "Elastic" in self.process_type and self.threshold > 0:
            raise RuntimeError(
                "Positive non-zero threshold passed for elastic cross section.")

        if nsamples == None and "dense" in grid_type:
            nsamples = self.nsamples
        if grid_type == None:
            grid_type == self.grid_type

        # Convert to min energy required by reaction (i.e. 0 if the threshold
        # is negative)
        B = max(0, self.threshold)
        if e_points:
            # Convert to numpy array
            e_points = np.array(e_points)
        else:
            # Create a energy points from preset grid based on the threshold.
            if grid_type == "log dense":
                e_points = np.logspace(self.SAMPLE_MIN,
                        np.log10(10**self.SAMPLE_MAX - B), nsamples) + B
            elif grid_type == "simple":
                if B > 0:
                    e_points = np.array([B, B+10**self.SAMPLE_MIN,
                        10**self.SAMPLE_MAX])
                else:
                    e_points = np.array([0, 10**self.SAMPLE_MAX])

        # Call function over grid (make sure to interpret as float)
        cross_section = vec(self.cs_func, otypes=[np.float64])(e_points)
        self.data = pd.DataFrame(np.stack((e_points, cross_section)).T,
                columns=["Energy (eV)", "Cross Section (m^2)"])
        # Sort data by energy
        self.data = self.data.sort_values(
            "Energy (eV)").reset_index(drop=True)
        # Clean up table near threshold
        self.zero_below_thresh()

    def zero_below_thresh(self):
        """Enforce the ThunderBoltz required cross section format.
        For processes with a non-zero threshold, include zero valued
        points at 0 eV and at threshold energy.

        Raises:
            RuntimeError: if there is no cross section data to format.
        """
        if self.data is None:
            raise RuntimeError(
                "Tabulated cross section data is not available for adjustment.")
        # Aliases for column names and threshold energy
        en = "Energy (eV)"
        csn = "Cross Section (m^2)"
        B = self.threshold

        if "Elastic" in self.process_type or B <= 0:
            # Extrapolate lowest cross section point to zero, if not available already
            if self.data.values[0,0] > 0:
                csv = self.data.values[0,1]
                self.data = pd.concat(
                    (pd.DataFrame([{en:0,csn:csv}]), self.data),
                    ignore_index=True)
            return

        # Otherwise, remove sampled data at and below the threshold
        self.data = self.data.loc[self.data[en] > B].copy()
        # Then add points at 0 eV and the threshold
        self.data = pd.concat((pd.DataFrame({en:[0,B], csn:[0,0]}),
            self.data), ignore_index=True)

    def require_cs(self):
        """Ensure there is some kind of cross section data associated with this
        process.

        Raises:
            RuntimeError: if there is no cross section data available.
        """
        if self.data is None:
            if not callable(self.cs_func):
                raise RuntimeError(
                    f"No cross section data available for process {self.name}")
            # Sample from analytic formula
            self.auto_sample()

    def to_cs_frame(self, a):
        """Convert any kind of two-dimensional data to a pandas
        DataFrame with columns `Energy (eV)` and `Cross Section (m^2)`
        """
        cols = ["Energy (eV)", "Cross Section (m^2)"]
        if isinstance(a, pd.DataFrame):
            return pd.DataFrame(a.values, columns=cols)
        return pd.DataFrame(a, columns=cols)

    def auto_sample(self):
        """Check if the cross section needs a dense grid or not by comparing
        simple and dense grids."""
        # Only need to check a few samples
        self.sample_cs(grid_type="log dense", nsamples=10)
        dense = copy.deepcopy(self.data)
        self.sample_cs(grid_type="simple")
        simple = copy.deepcopy(self.data)
        # Check difference through linear interpolation.
        cols = ["Energy (eV)", "Cross Section (m^2)"]
        dx, dy = dense[cols].values.T
        sx, sy = simple[cols].values.T
        s_interp = np.interp(dx, sx, sy)
        # Divide by zero mask
        msk = dy != 0
        is_diff = np.sum(np.abs((s_interp - dy)[msk]/dy[msk])) > 1e-6
        # Data is simple here, but if there is a discrepancy, use dense data
        if is_diff:
            self.sample_cs(grid_type="log dense")

    @property
    def nsamples(self):
        return self._nsamples

    @property
    def cs_func(self):
        return self._cs_func

    @nsamples.setter
    def nsamples(self, ns):
        self._nsamples = ns
        # Re-sample cross section if there is a callable cross section function
        if callable(self.cs_func):
            self.auto_sample()

    @cs_func.setter
    def cs_func(self, f):
        self._cs_func = f
        if callable(f):
            self.auto_sample()



import copy
import json
import os
from os.path import join as pjoin
from os import listdir as ls
import re
from typing import Tuple

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.constants import physical_constants as pc

from thunderboltz import parsing
from thunderboltz import data_path
from thunderboltz.kinetic import Process
from thunderboltz.plotting import styles

# Physical Constants
ME = 9.1093837e-31 # electron mass, kg
QE = 1.60217663e-19 # elementary charge C or J/eV
AMU_TO_KG = 1.6605e-27 # kg / amu
a_0 = 5.29177e-11 # Bohr radius in m
eVRyd = 13.6056980659 # eV per Rydberg

# Map LXCat process types to ThunderBoltz process types.
lxcat_map = {
    "ELASTIC": "Elastic",
    "EFFECTIVE": "Elastic",
    "ROTATIONAL": "Inelastic",
    "EXCITATION": "Inelastic",
    "IONIZATION": "Ionization",
    # "ATTACHMENT": Not yet implemented in ThunderBoltz
}

class CrossSections(object):
    """ThunderBoltz cross section set data type. Consists of
    a set of cross sections each with a file reference and a
    reaction table.

    Args:
        directory (str): The path to a ThunderBoltz simulation
            directory in which input files are to be written.
            Default is ``None``.
        input_path (str): The path to a set of ThunderBoltz
            input files from which input data can be read.
            The file structure should be something like:
            ::
               path/to/input_path
                |———indeck_file.in  <— The main ThunderBoltz indeck file.
                |———cross_sections  <— Cross section directory
                |   |———cs1.dat     <— ThunderBoltz-formatted cross section file.
                |   |———cs2.dat     <— ThunderBoltz-formatted cross section file.
                |   ...
                ...

        cs_dir_name (str): The name of the cross section directory.
        input_fname (str): The name of the main ThunderBoltz indeck
            file. Default is None, in which case the indeck will be
            searched for in `input_path` and must end with `.in`.
    """
    # TODO: Update this API so that cross section files and indeck files may be
    # separate if desired, but defaults to finding first compatible .in file and
    # cross_section directories

    def __init__(self, directory=None, input_path=None, read_cs_data=True,
            cs_dir_name="cross_sections", input_fname=None):
        #: Place for cs files in simulation dir
        self.cs_dir_name = cs_dir_name
        #: Input deck filename default
        self.input_fname = input_fname

        # Store data
        cols = ["csfile", "r1", "r2", "rtype", "B", "p1", "p2", "model_name", "params"]
        types = [str]+[np.int64]*2+[str]+[np.float64]+[np.int64]*2+[str]*2
        #: The reaction table, with columns
        self.table = pd.DataFrame({k: pd.Series([], dtype=dtype)
            for k, dtype in zip(cols, types)})

        #: Data tables for each cross section.
        self.data = {} # Store the cross section data.
        # Hyper parameters
        self.input_path = input_path #: Input path to default input data
        self.def_cs_dir = None
        self.def_indeck = None
        if input_path:
            # Read input from default cs data files
            self.read(input_path, read_cs_data=read_cs_data)
        # Otherwise, do not read/store any CS data

        if directory:
            self.find_infile()

            if input_path:
                # Create cs_dir for new calculation
                self.indeck = pjoin(directory, self.input_fname)
                self.cs_dir = pjoin(directory, self.cs_dir_name)
                # Make dir.
                if not os.path.isdir(self.cs_dir):
                    os.mkdir(self.cs_dir)

    def find_infile(self):
        """Look for indeck file in input_path.

        Raises:
            RuntimeError: If input_path is not set, if multiple indecks
            are found, or if no indecks are found.
        """
        if not self.input_path:
            raise RuntimeError("No input directory to search for indeck file in.")
        if self.input_fname is None:
            # Try to find one on type ".in"
            fs = [f for f in ls(self.input_path) if f.endswith(".in")]
            if not len(fs):
                raise RuntimeError("No input file found in indeck directory.")
            if len(fs) > 1:
                raise RuntimeError("Multiple input files found in indeck directory")
            self.input_fname = fs[0]

    def read(self, input_path, read_cs_data=True):
        """Read ThunderBoltz cross section data from a directory with a
        single input file and a set of cross section files.

        Args:
            input_path (str): The path to the directory
                with cross section data. If specified, CrossSections.input_path
                will be updated as well.
            read_cs_data (bool): If ``False``, only read the process header
                information, and not the actual cs data itself, default
                is ``True``.
        """
        # Make this input path as the reference input directory for this object
        self.input_path = input_path
        # Locate a '.in' file.
        self.find_infile()
        self.def_cs_dir = pjoin(input_path, self.cs_dir_name)
        self.def_indeck = pjoin(input_path, self.input_fname)

        # Reinitialize the reaction table
        cols = ["csfile", "r1", "r2", "rtype", "B", "p1", "p2", "model_name", "params"]
        types = [str]+[np.int64]*2+[str]+[np.float64]+[np.int64]*2+[str]*2
        self.table = pd.DataFrame({k: pd.Series([], dtype=dtype)
            for k, dtype in zip(cols, types)})
        # Read the input deck for cross section defaults.
        with open(self.def_indeck, "r") as f:
            for line in f.readlines():
                if "CS" not in line or "CC" in line:
                    continue
                l = line.split()[1:len(cols)]
                # set csfile relative to cs_dir
                l[0] = "/".join(l[0].replace("./", "").split("/")[-1:])
                lps = " ".join(line.split()[len(cols):])
                # Save to a data table
                items = l + [lps]
                row = {k: v for k, v in zip(cols, items)}
                self.table = pd.concat(
                    (self.table, pd.DataFrame([row])), ignore_index=True)
        # Recast types (they were read in as strings)
        self.table = self.table.astype({k: v for k, v in zip(cols, types)})

        if not read_cs_data:
            return

        # Read the input cross section files.
        for csfile in self.table.csfile.unique():
            with open(pjoin(self.def_cs_dir, csfile), "r") as f:
                # Loop through cross section data
                df = pd.DataFrame()
                for i, line in enumerate(f.readlines()):
                    e, cs = list(map(float, line.split()))
                    row = {"Energy (eV)": e, "Cross Section (m^2)": cs}
                    df = pd.concat((df, pd.DataFrame([row])), ignore_index=True)
                # Save data tables into a hashmap
                self.data[csfile] = df


    def write(self, directory=None):
        """Write cross section files into the simulation
        cross section directory.

        Args:
            directory (str): Option to write to a specific directory.
                If provided, ``CrossSections.cs_dir`` will be updated.
                Default is None.
        Raises:
            RuntimeError: If no directory is set or provided.
        """
        if directory:
            self.cs_dir = pjoin(directory, self.cs_dir_name)
        if not self.cs_dir:
            raise RuntimeError(
                "Nowhere to write, specify directory")
        for cs_file, cs_data in self.data.items():
            # cs_file = reaction + ".dat"
            cs_path = pjoin(self.cs_dir, cs_file.replace("./", ""))
            cs_dir = os.path.split(cs_path)[0]
            if not os.path.isdir(cs_dir):
                os.makedirs(cs_dir)
            with open(cs_path, "w") as f:
                f.write("\n".join("\t".join(str(s) for s in pair) for pair in
                    cs_data.values))

    def add_process(self, p):
        """Take a Process object and update the cross
        section data and cross section reaction table.

        Args:
            p (Process): The process object for a single
                type of interaction.
        """
        # Make sure cross section data is available for this process
        p.require_cs()
        cs_row = p.to_df()
        name = cs_row.csfile.values[0]
        cs_dat = copy.deepcopy(p.data)
        # Add descriptors to cross section table
        self.table = pd.concat(
            (self.table, cs_row), ignore_index=True)
        # Specify cross section data
        self.data[name] = cs_dat

    def add_processes(self, ps):
        """Add multiple cross sections to the reaction table.

        Args:
            ps (list[Process]): A list of process objects to add.
        """
        for p in ps: self.add_process(p)

    def add_differential_model(self, rtype, name, params=None):
        """Add a differential model to a certain type of process.

        Args:
            rtype (str): "Elastic", "Inelastic", or "Ionization",
                the broad collision process type.
            name (str): The name of the differential process model.
                Available built-in options for each ``rtype`` are:

                ========== ====================
                ``rtype``  ``name``
                ========== ====================
                Elastic    Park, Murphy
                Ionization Equal, Uniform
                ========== ====================

            params (list[float]): Optional list of parameters required
                by the differential model.
        """
        msk = self.table.rtype.str.contains(rtype)
        self.table.loc[msk, "model_name"] = name
        if params:
            self.table.loc[msk, "params"] = " ".join(str(p) for p in params)

    def set_fixed_background(self, fixed=True):
        """Set all particle conserving processes to have the
        `FixedParticle2` tag or not."""
        c = self.table
        # Make dynamic fist
        c.loc[:,"rtype"] = c.rtype.str.replace("FixedParticle2", "")
        if fixed:
            # Add fixed tags to non ionizing collisions.
            nion = ~c.rtype.str.contains("Ionization")
            c.loc[nion,"rtype"] += "FixedParticle2"

    def plot_cs(self, ax=None, legend=True, vsig=False,
                thresholds=False, **plot_args):
        r"""Plot the cross sections models.

        Args:
            ax (:class:`~.matplotlib.axes.Axes` or None): Optional
                axes object to plot on top of, default is ``None``.
                If ax is ``None``, then a new figure and Axes object
                will be created.
            legend (bool or dict): Activate axes legend if true,
                default is True. If a dictionary is passed, it is
                interpreted as arguments to :meth:`~.matplotlib.axes.Axes.legend`.
            vsig (bool): Plot :math:`\sqrt{\frac{2\epsilon}{m_{\rm e}}}\sigma(\epsilon)`
                rather than :math:`\sigma(\epsilon)` on the y-axis.
            thresholds (bool): if ``True``, plot
                :math:`\frac{\epsilon}{\epsilon_{\rm ion}} - 1`
                rather than :math:`\epsilon` on the x-axis.
            **plot_args: Optional arguments passed to Axes.plot().
        Returns:
            ax (matplotlib.axes.Axes): The axes object.
        """
        # Column name aliases
        en = "Energy (eV)"
        csn = "Cross Section (m^2)"

        # Generate axes object if not provided
        if ax is None:
            _, ax = plt.subplots(figsize=(12, 10))
        # Loop through the processes
        for (csfile, B), color in zip(
                self.table[["csfile", "B"]].values, styles.DCC):
            pname = csfile.replace(".dat", "")
            cs_dat = self.data[csfile]
            x = cs_dat[en]/B - 1 if thresholds else cs_dat[en]
            cs = (np.sqrt(2*cs_dat[en]*QE/ME)*cs_dat[csn]
                  if vsig else cs_dat[csn])
            ax.plot(x, cs, label=pname, c=color)

        # Setup axes settings
        ax.set_xscale("log")
        ax.set_yscale("log")
        ax.set_xlabel(r"$\epsilon$ (eV)")
        if thresholds:
            ax.set_xlabel(r"$\frac{\epsilon}{\epsilon_{\rm ion}} - 1$")
        ax.set_ylabel(r"$\sigma(\epsilon)$")
        if vsig:
            ax.set_ylabel(r"$\sqrt{\frac{2\epsilon}{m_{\rm e}}}\sigma(\epsilon)$")
        if legend:
            ax.legend(fontsize=12)
        if isinstance(legend, dict):
            lg_args = {"fontsize": 12}
            lg_args.update(legend)
            ax.legend(**lg_args)

        return ax

    def get_deck(self):
        """Return the string formatted cross section table portion of the
        ThunderBoltz indeck."""
        s = ""
        for row in self.table.values:
            cspath = pjoin(self.cs_dir_name, row[0])
            description = " ".join(str(a) for a in row[1:] if not pd.isna(a)).rstrip() + "\n"
            s += f"CS {cspath} {description}"
        return s

    def from_LXCat(self, fname):
        """Load cross section data from an LXCat .txt file"""
        if not os.path.exists(fname):
            raise IOError(f"LXCat file {fname} does not exist.")
        if self.data or len(self.table):
            raise RuntimeError(f"LXCat file was passed when cross section "
                                "data already exists")
        df = parsing.read_LXCat_cs(fname)
        # Loop through each process
        for process_string, pdf in df.groupby("process_string"):
            # Derive the names of each process from the process string
            name = process_string.replace(" ", "")
            # TODO: implement LXCat species interpretation for TB species indexing
            # parse_process_string(name)
            # Make sure the threshold matches
            if not len(pdf.threshold.unique()) == 1:
                raise RuntimeError("Process string does not uniquely identify process")
            cs_dat = (pdf[["Energy (eV)", "Cross section (m2)"]].copy()
                .sort_values("Energy (eV)")
                .reset_index(drop=True)
                .rename(columns={"Cross section (m2)": "Cross Section (m^2)"})
            )
            # Create thunderboltz.kinetic.Process to enfore threshold rules
            ptype = pdf.process_type.unique()[0]
            B = pdf.threshold.unique()[0]
            B = 0.0 if pd.isna(B) else B
            if ptype not in lxcat_map:
                raise NotImplementedError(f"There is no current {ptype} "
                        "process implemented in ThunderBoltz")
            self.add_process(Process(lxcat_map[ptype],
                0, 1, 0, 1, threshold=B, cs_data=cs_dat, name=name))

        return self # So it's easier to load directly from LXCat

### Automated cross section constructions.
def He_TB(n=4, egen=True, analytic_cs=True, eadf="default", ECS=None,
          nsamples=250, mix_thresh=300., fixed_background=True):
    """Generate parameterized He cross section sets in the ThunderBoltz format.
    The data is from Igor Bray and Dmitry V Fursa 2011
    J. Phys. B: At. Mol. Opt. Phys. 44 061001.

    Args:
        n (int): Include CCC excitation processes from the ground state to
            (up to and including) states with principle quantum number ``n``.
        egen (bool): Allow secondary electron generation for the ionization model.
        analytic_cs (str or bool): use either tabulated data, analytic fits,
            or a mix of both. Options are ``False``, ``True``, or ``"mixed"``.
        eadf (str) Elastic angular distribution function model. Options are
            ``"default"``, or ``"He_Park"``.
        ECS (str or None): The total elastic cross section model. Options are
            ``"ICS"`` or ``"MTCS"``, default is ``"ICS"`` if an anisotropic
            angular distribution function is used and ``"MTCS"`` if an isotropic
            angular distribution function is used.
        nsamples (int): The number of tabulated cross section values for analytic
            sampling.
        mix_thresh (float): If ``analytic_cs`` is ``"mixed"``, use numerical data
            at energies lower than this threshold value (in eV), and use analytic
            data at higher energies.
        fixed_background (bool): Flag to append "FixedParticle2" to each of the
            reaction types in the indeck.

    Returns:
        Tuple[dict,dict]: The :class:`~.thunderboltz.CrossSections` object for
            Helium and the dictionary of ThunderBoltz parameters suitable
            for the cross section model.
    """

    # Determine elastic flag
    elastic_only = False
    if n==0:
        elastic_only = True
        n = 1
    # Select elastic integrated cross section type
    if ECS is None:
        ECS = "MTCS" if eadf == "default" else "ICS"
    # Load in cross sections from CCC data
    cmod = He_CCC(n, ECS=ECS)
    if analytic_cs:
        # Use continuous fits
        cmod_aly = make_He_aly_cs(n, ECS, nsamples=nsamples)
        if analytic_cs == "mixed":
            cmod_aly = split_cs_model(cmod, cmod_aly, mix_thresh, ECS)
        # overwrite original mod
        cmod = cmod_aly

    # Make sure threshold format is correct
    cmod = fix_thresh(cmod)

    # Create CS object from cmod dict
    itype = "Ionization" if egen else "IonizationNoEGen"
    tbcs = mod_to_tbcs(cmod, ion_type=itype, elastic_only=elastic_only)
    tbcs.set_fixed_background(fixed_background)

    # Specific TB settings for He background gas
    # These cannot be overwritten
    tbp = {
        "SP": 2,
        "NP": [100000, 10000],
        "TP": [0.0, 0.0259],
        "VV": [0.0, 0.0],
        "QP": [-1.0, 0.0],
        "MP": [5.4857e-4, 4.0],
        "FV": [1000, 150000, 0],
    }

    return tbcs, tbp

# Helper functions
def mod_to_tbcs(mod, ion_type="Ionization", elastic_only=False):
    """Convert dictionary cross section format to CrossSections Object."""
    sind = {"r1": 0, "r2": 1, "p1": 0, "p2": 1}
    # Elastic CS, assume fixed particles
    cmod = mod["cross_sections"]
    elastic_type = cmod["elastic"].columns[1]
    # cs_data init with elastic data
    df = pd.DataFrame([dict(csfile=f"{elastic_type}.dat", B=0.,
        rtype="ElasticFixedParticle2", **sind)])
    if not elastic_only:
        # Excitation cross sections
        for exc, B in mod["exc_thresh"].items():
            dfrow = pd.DataFrame([dict(csfile=f"{exc}.dat", B=B,
                rtype="InelasticFixedParticle2", **sind)])
            df = pd.concat((df, dfrow), ignore_index=True)
        # Ionization cross section
        B = mod["ion_thresh"]
        ion_row = pd.DataFrame([dict(csfile="ionz.dat", B=B, rtype=ion_type, **sind)])
        df = pd.concat((df, ion_row), ignore_index=True)

    # Convert type
    df = df.astype({"r1": int, "r2": int, "B": float, "p1": int, "p2": int})
    # Reorder columns
    df = df[["csfile", "r1", "r2", "rtype", "B", "p1", "p2"]].copy()

    # Create cs data dictionary
    # Elastic data
    cs_dat = {f"{elastic_type}.dat": cmod["elastic"].rename(
        columns={elastic_type: "Cross Section (m^2)"})}

    # Excitation data
    if not elastic_only:
        for name, exc_dat in cmod.items():
            if name in ["elastic", "ionization"]:
                continue
            c2 = exc_dat.columns[1]
            cs_dat.update({f"{name}.dat": exc_dat.rename(
                columns={c2: "Cross Section (m^2)"})})

        # Ionization data
        cs_dat.update({f"ionz.dat": cmod["ionization"].rename(
            columns={"TICS (m^2)": "Cross Section (m^2)"})})

    # Create CrossSections object
    cs = CrossSections()
    cs.table = pd.concat((cs.table, df), ignore_index=True)
    cs.data = cs_dat
    return cs

def He_CCC(nmax=3, ECS="MTCS"):
    """Helium package cross section model from CCC data.
    Includes excitation cross sections up to nmax"""
    # Create tables
    # Import excitation cross sections and threshold energies.
    ccc_path = pjoin(data_path, "CCC")
    tcs = pd.read_csv(pjoin(ccc_path, "He_DCS_tcs.csv"))
    potl = pd.read_csv(pjoin(ccc_path, "He_DCS_potl.csv"))

    # Manipulate extract for bolsig.
    potl = potl[["Energy (eV)"] + [c for c in
        potl.columns if "(cm^2)" in c]]
    potl = convert(potl, "cm^2", "m^2", drop=True)
    tcs = convert(tcs, "cm^2", "m^2")

    # Threshold energy from sample DCS file.
    be = pd.read_csv(pjoin(ccc_path, "He_thresholds.csv"), header=None,
            names=["transition", "Energy (Ryd)"])
    be = convert(be, "Ryd", "eV")

    # ELASTIC
    eltable = potl[["Energy (eV)", ECS]]
    # Calculate electron-target mass ratio
    He_m = 2*(pc["electron mass"][0]) + pc["alpha particle mass"][0]
    etmr = pc["electron mass"][0] / He_m

    # IONIZATION
    iotable = tcs[["Energy (eV)", "TICS (m^2)"]]
    io_thr = -be.loc[be.transition=="s1S", "Energy (eV)"].values[0]

    # EXCITATION
    # Select the desired excitation states.
    n_states = limit_excitations(potl, nmax)
    n_thresholds = exthresh_from_BE(be, n_states)
    # Binding energies obtained from CCC data: Igor Bray and Dmitry V Fursa 2011 J. Phys. B: At. Mol. Opt. Phys. 44 061001

    # Assemble and name cross section tables.
    cs_tables = {
        "elastic": eltable,
        "ionization": iotable,
    }
    cs_tables.update({k: potl[["Energy (eV)", k]]
        for k in n_states})

    # Assemble model with all info needed for BOLSIG.
    model = {
        "format": "TABULAR", # Indicate that cross sections are stored as DataFrames.
        "name": "He_CCC_n<={}".format(nmax),
        "cross_sections": cs_tables,  # Columns of energy, CS for every interaction.
        "exc_thresh": n_thresholds, # threshold for all inelastic processes.
        "ion_thresh": io_thr, # eV
        "etmr": etmr,
    }
    return model

def elastic_mtcs_LANL(e, sigt):
    """Takes incident energy and ICS(e) function and returns the
    momentum transfer cross section for He.

    This functional form is obtained from Park et al.
    https://iopscience.iop.org/article/10.1088/1361-6595/ac781f
    """
    ps = [0.283, 0.667, 0.0307, 16.971, 6.59, 29.02, 0.0258, 0.295, 0.00328, 1.794]
    a, laf, lab = ps[:4], ps[4:7], ps[7:]
    # Derived parameters
    C = a[0] + a[1]*np.tanh(a[2]*(e - a[3]))
    etaF = (laf[0] / (e + laf[1])) + laf[2]
    etaB = (lab[0] / (e + lab[1])) + lab[2]
    lF = (etaF * (etaF + 1)) / np.pi
    lB = (etaB * (etaB + 1)) / np.pi
    mF = (2*etaF + 1)
    mB = (2*etaB + 1)
    # Evaluation
    Phi = lambda l, m, n: ((l/(n**2)) * (-1 + ((m+n)/(m-n)) + np.log((m-n)/(m+n))))
    sigm = 2*np.pi*sigt*(C*Phi(lF, mF, -1) + (1-C)*(Phi(lB, mB, 1)))
    return sigm

def elastic_ics_LANL(e):
    """Takes incident energy and computes ICS (m^2) using fitting function.

    This functional form is obtained from Park et al.
    https://iopscience.iop.org/article/10.1088/1361-6595/ac781f
    """
    b = [1.47183e-18, 6.23206e-18, -1.30586e-18, 7.52062e-1,
         1.92954e-1, 2.42524, 4.49405, 1.25050e2]
    t1 = b[0] + (b[1]/(e+1)) + (b[2]/(e**2 + 1))
    t2 = (e**b[3] + np.exp(-b[4]*e))**b[5]
    t3 = (e+1)**3 + b[6]*(e+1)**2 + b[7]*(e+1)
    return t1*(t2/t3)

def fix_thresh(mod):
    """Fix the threshold value such that cross sections are 0
    at their threshold."""

    en = "Energy (eV)"
    csn = "cross_sections"
    el = mod[csn]["elastic"]
    ECS = el.columns[1]
    # Check elastic cross section to see if it has data at 0 eV.
    if el[en].values[0] != 0.:
        mod[csn]["elastic"] = pd.concat((pd.DataFrame({en: [0.], ECS:
            el.iloc[[0], 1]}), el), ignore_index=True)
    # Fix thresh on ionization cross section.
    ionn = "TICS (m^2)"
    B = mod["ion_thresh"]
    ion = mod[csn]["ionization"]
    mod[csn]["ionization"] = pd.concat((pd.DataFrame({en: [0.0,
        B], ionn: [0.0, 0.0]}), ion[ion[en] > B]), ignore_index=True)

    # Fix thresh on excitation cross sections.
    for proc in mod[csn]:
        if proc in ["elastic", "ionization"]:
            continue
        thr = mod["exc_thresh"][proc]
        inel = mod[csn][proc]
        mod[csn][proc] = pd.concat((pd.DataFrame({en: [0.0,
            thr], proc: [0.0, 0.0]}), inel[inel[en] > thr]), ignore_index=True)
    return mod


def make_He_aly_cs(n=3, ECS="ICS", nsamples=248):
    """Construct a grid from analytic He cross section fits proposed by
    Ralchenko et al. https://doi.org/10.1016/j.adt.2007.11.003."""

    # Get thresholds
    mod = He_CCC(n, ECS=ECS)
    exth = mod["exc_thresh"]

    # Named energy column
    en = "Energy (eV)"
    ionn = "TICS (m^2)"

    # Read in cs parameters from Ralchenko
    csp = pd.read_csv(pjoin(data_path, "Ralchenko_He_CS.csv"))
    # Only use transitions from ground state
    csp = csp[csp.i == "s1S"].copy()

    def sample_exc(name, e):
        """Sample excitation cross section with 'name' at energy 'e' (thresholds)"""
        # params
        rp = csp[csp.f == name].copy()
        a = rp.loc[:,"A1":"A6"].values[0]
        # Employ Ralchenko model
        if rp.table.values[0] == "dipole-allowed":
            return (a[0]*np.log(e) + a[1] + a[2]/e + a[3]/e**2 + a[4]/e**3)*(e + 1)/(e+a[5])
        elif rp.table.values[0] == "dipole-forbidden":
            return (a[0] + a[1]/e + a[2]/e**2 + a[3]/e**3) * e**2/(e**2 +a[4])
        elif rp.table.values[0] == "spin-forbidden":
            return (a[0] + a[1]/e + a[2]/e**2 + a[3]/e**3) / (e**2 + a[4])

    def sample_exc_set(name):
        """Sample a set of energies (eV) of excitation cross section with transition 'name'"""
        # Get threshold energy
        B = exth[name]
        # Make energies in eV
        es = np.logspace(-4, np.log10(10**6 - B), nsamples) + B
        # Calculate omega with energy in thresholds
        omega = np.vectorize(lambda e: sample_exc(name, e))(es/B)
        # Add pre factor to convert to cross section in m^2
        prefac = np.pi*a_0**2*(eVRyd/es)
        # Create DataFrame for this data
        dat = pd.DataFrame({en: es, name: prefac*omega})
        # Add extrapolation near 0 energy
        return pd.concat((pd.DataFrame({en: [0, B], name: [0, 0]}),
                          dat[dat[en] > B]), ignore_index=True)

    # Update elastic model
    # Add elastic model, include 0 eV data point
    es = np.concatenate(([0], np.logspace(-4, 6, nsamples)))
    el_df = pd.DataFrame({en: es, ECS: elastic_ics_LANL(es)})
    if ECS=="MTCS":
        # Use ICS to compute MTCS from LANL model
        el_df[ECS] = elastic_mtcs_LANL(es, el_df[ECS])
    # Add one point at 0 incident energy with the same value of cross section as first NZ point
    mod["cross_sections"]["elastic"] = el_df

    # Update ionization model
    a = csp.loc[csp.table=="ionization", "A1":"A6"].values[0]
    # Make a[1] = A1 for clarity
    a = np.concatenate(([0], a))
    B = mod["ion_thresh"]
    # Do not include 0 eV data point
    es = np.logspace(-4, np.log10(10**6 - B), nsamples) + B
    TICS = (1e-17/(B*es))*(a[1]*np.log(es/B) + sum(a[i]*(1-B/es)**(i-1) for i in range(2,7)))
    ion_df = pd.DataFrame({en: es, ionn: TICS})
    # Add a point at 0,0 and
    mod["cross_sections"]["ionization"] = pd.concat((
        pd.DataFrame({en: [0, B], ionn: [0, 0]}), ion_df[ion_df[en]>B]), ignore_index=True)

    # Update excitation cross sections.
    for name in mod["cross_sections"]:
        if name in ["ionization", "elastic"]:
            continue
        mod["cross_sections"][name] = sample_exc_set(name)

    return mod

def split_cs_model(m1, m2_, thresh, ECS):
    """Split two cross section models such that `m1` provides energies lower
    than `thresh` and `m2` provides energies higher than `thresh`."""
    en = "Energy (eV)"
    # Copy right mod
    m2 = copy.deepcopy(m2_)

    def split_cs(key):
        col = key
        if key == "elastic":
            col = ECS
        elif key == "ionization":
            col = "TICS (m^2)"
        m1e = m1["cross_sections"][key]
        m1cut = m1e[m1e[en] < thresh].copy()
        m2e = m2["cross_sections"][key]
        m2cut = m2e[m2e[en] > thresh].copy()
        m2["cross_sections"][key] = pd.concat((m1cut, m2cut),
                ignore_index=False).reset_index(drop=True)
    # Swap elastic model
    split_cs("elastic")
    split_cs("ionization")
    for proc in m2["cross_sections"]:
        if proc in ["elastic", "ionization"]:
            continue
        split_cs(proc)

    return m2

def parse_process_string(s):
    """Find species stings from process string."""
    # remove unnecessary eV tags
    s = re.sub(r"\([^()]*eV\)", "", s)

def limit_excitations(df, nmax=3):
    """Ignore transitions with principle quantum numbers
    above `nmax`."""
    # Isolate relevant excitations from CCC He data
    complete_states = [c for c in df.columns
        if not df[c].isnull().values.all()]
    # singlet and triplet states.
    st_states = [s for s in complete_states if s[0] in ["s", "t"]]
    # Quantum number of allowed transitions.
    n = np.arange(2, nmax+1)
    n_states = [s for s in st_states if int(s[1]) in n]
    return n_states

def exthresh_from_BE(be, states):
    """Create mapping from state transition name to excitation threshold"""
    smap = lambda s: s in states
    gs_be = be.loc[be.transition=="s1S", "Energy (eV)"].values
    exthresh = be[be.transition.map(smap)].copy()
    exthresh.loc[:, "thresh"] = exthresh["Energy (eV)"] - gs_be
    return {k: v for k, v in exthresh[["transition", "thresh"]].values}

def convert(df, u1, u2, inv=False, drop=False, add=False):
    """Convert easily between units with a labeled DataFrame.
    Columns in the format `<name> (<unit>)` will be converted."""
    umap = {
        # Area in one m^2
        "m^2": 1,
        "m2": 1,
        "Angs^2": 1e20,
        "10^-22m^2": 10e22,
        "cm^2": 100**2,
        "10^-16cm^2": 1e16*100**2,
        "a_0^2": 1.8897259e10**2,
        "Mb": 1e22,
        "kb": 1e25,
        # Energy in one eV
        "eV": 1.,
        "Ryd": 1/13.6056980659,
    }
    for col in df.columns:
        if u1 in col:
            if drop:
                # Convert but remove the unit tag.
                ncol = col.replace(" "+col.split()[1], "")
            else:
                # Change unit tag to new unit
                ncol = col.replace(u1, u2)
            if inv:
                # Convert inverse units
                conversion = umap[u1]/umap[u2]
            else:
                # Convert units
                conversion = umap[u2]/umap[u1]
            if add:
                # Make a new column
                df.loc[:, ncol] = df[col] * conversion
            else:
                # Rename the old column
                df.loc[:, col] *= conversion
                df = df.rename(columns={col: ncol})
    return df



from collections import OrderedDict
import copy
import itertools
import os
from os.path import join as pjoin
from pathlib import Path
import sys

import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
from numpy import vectorize as vec
import seaborn as sns
import scipy

from thunderboltz import parameters
from thunderboltz import parsing
from thunderboltz.plotting import pdplot
from thunderboltz.plotting import styles

ME = 9.1093837e-31 # kg
QE = 1.60217663e-19 # C or J/eV
AMU_TO_KG = 1.6605e-27 # kg / amu

def plot_timeseries(fig, ts, series=["MEe", "mobN", "a_n"], save=None):
    """Plot ThunderBoltz output values as function of time."""
    if len(fig.axes) > 1:
        # Reset the twin axes
        [a.remove() for a in fig.axes[1:]]
    # Create twin axes
    [fig.axes[0].twinx() for param in series[1:]]
    ax = fig.axes[0]
    ax.grid(which='major', color='gray', linestyle='-', alpha=0.5)
    ax.grid(which='minor', color='gray', linestyle='-', alpha=0.3)
    # Plot first dataset on left hand side
    ax.plot(ts.t, ts[series[0]], label=styles.label_maps(series[0]), c="green")
    # Get lims for before plotting verticle lines
    ylm = ax.get_ylim()
    ax.set_ylim(*ylm)
    ax.set_ylabel(styles.label_maps(series[0]))
    ax.set_xlabel("Time (s)")
    colors = ["purple", "red", "orange", "pink", "green", "blue", "#afa", "#33b"]
    offset = 1.
    for i, (s, c) in enumerate(zip(series[1:], colors)):
        axtwin = fig.axes[i+1]
        axtwin.spines.right.set_position(("axes", offset))
        offset += 0.18
        axtwin.set_ylabel(styles.label_maps(s))
        axtwin.plot(ts.t, ts[s], c=c, label=styles.label_maps(s))
    handles, labels = figure_legend(fig.axes)
    leg = fig.legend(handles, labels, loc="lower center", ncol=100)
    fig.subplots_adjust(right=max(.2, 2-.9*offset), bottom=.18, top=.80)
    return fig

def plot_rates(ax, ts, bins=150):
    """Plot the reaction rates on the."""
    pass

def plot_eedfs(ax, esamples, bins=150):
    """Plot the electron energy density funciton"""
    pass

def plot_vdfs(vdfs, pair=["vx", "vy"], bins=50):
    """Plot the electron velocity distribution function."""
    # Seaborn color palette
    rocket = sns.color_palette("rocket_r", as_cmap=True)
    # Create joint plot
    jplot = sns.jointplot(x=vdfs[pair[0]]/1e6, y=vdfs[pair[1]]/1e6, kind="hist",
            cbar=True, marginal_ticks=False, cmap=rocket, bins=bins,
            marginal_kws={"color": (209/255, 38/255, 65/255), "bins": bins})
    plt.subplots_adjust(left=0.1, right=0.8, top=0.95, bottom=0.1)
    # Get the positions of the joint and marginal density plots
    jpos = jplot.ax_joint.get_position()
    xmpos = jplot.ax_marg_x.get_position()
    ympos = jplot.ax_marg_y.get_position()
    jplot.ax_joint.set_position([jpos.x0+.02, jpos.y0, xmpos.width, jpos.height])
    jplot.ax_marg_x.set_position([jpos.x0+.02, jpos.y1, xmpos.width, xmpos.height])
    # Move colorbar to the outside
    cbar = jplot.fig.axes[-1]
    cbar.set_position([.83, jpos.y0, .07, jpos.height])
    jax = jplot.fig.axes[0]
    jax.set_xlabel(f"$v_{pair[0][1]}$ (10$^6$ m/s)")
    jax.set_ylabel(f"$v_{pair[1][1]}$ (10$^6$ m/s)")
    jplot.fig.set_size_inches(10, 8.9)
    return jplot.fig

def align_vdfs_on(tbs, params, save=None, bins=100):
    """Align vdfs figures on every unique combination of `params`.

    tbs: list<ThunderBoltz>
        List of ThunderBoltz objects to create VDF plots for.
    params: list | str
        ThunderBoltz parameter key(s) that form groups to align scales on."""
    if isinstance(params, str):
        params = [params]
    def get_val(tb, p):
        tbp = tb.tb_params
        if p in tbp:
            return tbp[p]
        if p in tb.hp:
            return tb.hp[p]
    # Make table of unique sets
    param_table = pd.DataFrame([dict(tb=tb, **{p: get_val(tb, p) for p in params})
                  for tb in tbs])
    # Collect all the figure objects
    figs = []
    # Where to save them, if necessary
    directories = []
    for ids, tb_group in param_table.groupby(params):
        # Get directories
        directories = [tb.get_directory() for tb in tb_group["tb"]]
        # Plot and collect figures for each of the tb objects
        # TODO: implement bins for step grouping
        group_figs = [tb.plot_vdfs(steps="last", save=False, bins=bins)
                      for tb in tb_group["tb"].values]
        xy_figs = [f[0][0] for f in group_figs if f]
        xz_figs = [f[0][1] for f in group_figs if f]
        steps = [f[1][0] for f in group_figs if f]

        # If there are no figures available in this group, continue
        if not steps:
            continue

        # Get the axes and colorbar  limits for each plot
        xy_xlims = [f.axes[0].get_xlim() for f in xy_figs]
        xy_ylims = [f.axes[0].get_ylim() for f in xy_figs]
        xy_clims = [f.axes[0].collections[0].get_clim() for f in xy_figs]
        xz_xlims = [f.axes[0].get_xlim() for f in xz_figs]
        xz_ylims = [f.axes[0].get_ylim() for f in xz_figs]
        xz_clims = [f.axes[0].collections[0].get_clim() for f in xz_figs]

        # Get max lims within group for each pair
        xy_m_x = (min(l[0] for l in xy_xlims), max(l[1] for l in xy_xlims))
        xy_m_y = (min(l[0] for l in xy_ylims), max(l[1] for l in xy_ylims))
        xy_m_c = (min(l[0] for l in xy_clims), max(l[1] for l in xy_clims))
        xz_m_x = (min(l[0] for l in xz_xlims), max(l[1] for l in xz_xlims))
        xz_m_y = (min(l[0] for l in xz_ylims), max(l[1] for l in xz_ylims))
        xz_m_c = (min(l[0] for l in xz_clims), max(l[1] for l in xz_clims))

        # Set each of the lims the these extrema
        for fxy, fxz, step, directory in zip(xy_figs, xz_figs, steps, directories):
            figs.extend([fxy, fxz])
            fxy.axes[0].set_xlim(*xy_m_x)
            fxy.axes[0].set_ylim(*xy_m_y)
            fxy.axes[0].collections[0].set_clim(*xy_m_c)
            fxz.axes[0].set_xlim(*xz_m_x)
            fxz.axes[0].set_ylim(*xz_m_y)
            fxz.axes[0].collections[0].set_clim(*xz_m_c)
            if save:
                sname = Path(directory).parts[-1]
                if not os.path.isdir(save):
                    os.makedirs(save)
                saveas = pjoin(save, sname+f"_xy{step}.pdf")
                print(f"Saving VDF joint plot in {saveas}", flush=True)
                fxy.savefig(saveas)
                plt.close(fxy)
                saveas = pjoin(save, sname+f"_xz{step}.pdf")
                print(f"Saving VDF joint plot in {saveas}", flush=True)
                fxz.savefig(saveas)
                plt.close(fxz)
    return figs

def vdf_pct_err(tbs, params, benchmark, save=None, bins=100):
    """Calculate the % error of the velocity distribution density function.

    tbs: list<ThunderBoltz>
        List of ThunderBoltz objects to create VDF plots for.
    params: list | str
        ThunderBoltz parameters which are being compared in the error calculation.
    benchmark: list | str
        Value of `params` to compare all other calculations against. Must be
        same length as params.
    """

    if isinstance(params, str):
        params = [params]
    def get_val(tb, p):
        tbp = tb.tb_params
        if p in tbp:
            return tbp[p]
        if p in tb.hp:
            return tb.hp[p]
    # Make table of unique sets
    # print(*[dict(tb=tb, **{p: get_val(tb, p) for p in params})
                  # for tb in tbs], sep="\n\n")

    param_table = pd.DataFrame([dict(tb=tb, **{p: get_val(tb, p) for p in params})
                  for tb in tbs])
    # Collect all the figure objects
    figs = []
    # Where to save them, if necessary
    directories = []
    for ids, tb_group in param_table.groupby(params):
        # Get directories
        directories = [tb.get_directory() for tb in tb_group["tb"]]
        # Plot and collect figures for each of the tb objects
        # TODO: implement bins for step grouping
        group_figs = [tb.plot_vdfs(steps="last", save=False, bins=bins)
                      for tb in tb_group["tb"].values]
        xy_figs = [f[0][0] for f in group_figs]
        xz_figs = [f[0][1] for f in group_figs]
        steps = [f[1][0] for f in group_figs]

        # Set each of the lims the these extrema
        for fxy, fxz, step, directory in zip(xy_figs, xz_figs, steps, directories):
            figs.extend([fxy, fxz])
            if False:
                sname = Path(directory).parts[-1]
                if not os.path.isdir(save):
                    os.makedirs(save)
                saveas = pjoin(save, sname+f"_xy{step}.pdf")
                print(f"Saving VDF joint plot in {saveas}", flush=True)
                fxy.savefig(saveas)
                plt.close(fxy)
                saveas = pjoin(save, sname+f"_xz{step}.pdf")
                print(f"Saving VDF joint plot in {saveas}", flush=True)
                fxz.savefig(saveas)
                plt.close(fxz)
    return figs

def plot_bilog(x, y, ylog=False, xlabel=None, xlabel_left=None, ylabel=None, **plotting):
    """Make a log plot with two sides, positive x values are plotting on the
    right and negative xvalues are plotting on the left with a flipped axis.

    Args:
        x (Array-Like): The x coordinate data.
        y (Array-Like): The y coordinate data.
        ylog (bool): If set to ``True``, plot the y data on a log scale.
        xlabel (str): label both x axes.
        xlabel_left (str): Use a custom xlabel for the negative x-coordinate
            axes. Otherwise, use ``xlabel`` with ``"-"`` prepended.
        ylabel(str): label both y axes.
        **plotting: Keyword arguments that will be passed to
            :meth:`matplotlib.axes.Axes.plot`
    """

    # Separate data into positive ande negative
    x_pos = x[x>0]
    y_right = y[x>0]
    x_neg = -x[x<0]
    y_left = y[x<0]

    # Find limits based on data ranges
    xmin = np.min(np.abs(x))
    xmax = np.max(np.abs(x))*1.05
    yrange = np.max(y) - np.min(y)
    ymin = np.min(y) - 1.05*yrange
    ymax = np.max(y) + 1.05*yrange

    # Build figure
    fig, axs = plt.subplots(1, 2)
    fig.subplots_adjust(wspace=0)
    # axs[0].spines["right"].set(visible=False)
    axs[1].spines["left"].set(visible=False)
    # Remove overlapping labels
    axs[0].get_yticklabels()[-1].set(visible=False)
    axs[1].get_yticklabels()[0].set(visible=False)
    axs[1].yaxis.set_label_position("right")
    axs[1].yaxis.tick_right()
    # [lbl.set(visible=False) for lbl in axs[0].get_yticklabels()]
    # [lbl.set(visible=False) for lbl in axs[1].get_yticklabels()]
    # axs[1].set_yticks([])


    axs[0].set_xscale("log")
    axs[1].set_xscale("log")
    axs[0].set_xlim((xmin, xmax))
    axs[1].set_xlim((xmin, xmax))
    axs[0].set_ylim((ymin, ymax))
    axs[1].set_ylim((ymin, ymax))

    axs[0].grid(which='major', color='gray', linestyle='-', alpha=0.5)
    axs[0].grid(which='minor', color='gray', linestyle='-', alpha=0.3)
    axs[1].grid(which='major', color='gray', linestyle='-', alpha=0.5)
    axs[1].grid(which='minor', color='gray', linestyle='-', alpha=0.3)

    # Plot Data
    axs[0].plot(x_neg, y_left, **plotting)
    axs[1].plot(x_pos, y_right, **plotting)
    axs[0].invert_xaxis()

    if xlabel:
        axs[1].set_xlabel(xlabel)
        if xlabel_left:
            axs[0].set_xlabel(xlabel_left)
        else:
            axs[0].set_xlabel("-"+xlabel)
    if ylabel:
        axs[0].set_ylabel(ylabel)
        axs[1].set_ylabel(ylabel)

    return fig


def make_pair_plot(ax, vdfs, bins=150):
    """Plot the marginal and joint correlation distributions of the three
    velocity fields."""
    pass

def get_hashable(h):
    """Return a hashable string that describes the plotting style
    of a handle."""
    props = [h.get_ls(), h.get_lw(), h.get_c(),
        h.get_ms(), h.get_marker()]
    s = "".join(list(map(str, props)))
    return s

def figure_legend(axs, order=None):
    """Take series across all axes and combine them into one legend."""
    # Store unique handle representations.
    hs = set()
    # Map back to the actual line objects.
    lhmap = OrderedDict()
    for ax in axs:
        hand, lab = ax.get_legend_handles_labels()
        lm = {get_hashable(h)+"@"+l: h for l, h in zip(lab, hand)}
        lhmap.update(lm)
    labels = [l.split("@")[1] for l in lhmap.keys()]
    handles = list(lhmap.values())
    if order == None:
        return handles, labels
    return [handles[i] for i in [labels.index(j) for j in order]], order


